from typing import Callable, Dict, Optional

from stx.functions.built_in import resolve_code
from stx.functions.built_in import resolve_custom_style
from stx.functions.built_in import resolve_toc
from stx.functions.built_in import resolve_embed
from stx.functions.built_in import resolve_image
from stx.functions.built_in import resolve_warning
from stx.functions.built_in import resolve_admonition
from stx.functions.built_in import resolve_line_feed
from stx.functions.built_in import resolve_information
from stx.functions.built_in import layout_function

from stx.components import Component, FunctionCall

from stx.document import Document
from stx.utils.stx_error import StxError


FunctionType = Callable[[Document, FunctionCall], Component]


_functions: Dict[str, FunctionType] = {}


def add(function_key: str, function: FunctionType, override=False):
    if not override and function_key in _functions:
        raise StxError(f'Function is already registered: {function_key}')

    _functions[function_key] = function


def get(function_key: str) -> Optional[FunctionType]:
    if function_key not in _functions:
        return None
    return _functions[function_key]


# Built-in

add('img', resolve_image)
add('image', resolve_image)
add('code', resolve_code)
add('warning', resolve_warning)
add('info', resolve_information)
add('information', resolve_information)
add('admonition', resolve_admonition)
add('br', resolve_line_feed)
add('style', resolve_custom_style)
add('toc', resolve_toc)
add('embed', resolve_embed)
add('layout', layout_function)



import threading


class _ThreadContext:

    def __init__(self):
        self._local = threading.local()

    @property
    def parser(self):
        stack = self._get_parser_stack()

        if len(stack) == 0:
            return None

        return stack[-1]

    def _get_parser_stack(self):
        stack = getattr(self._local, 'parser_stack', None)

        if stack is None:
            stack = []
            setattr(self._local, 'parser_stack', stack)

        return stack

    def push_reader(self, reader):
        self._get_parser_stack().append(reader)

    def pop_reader(self):
        self._get_parser_stack().pop()


context = _ThreadContext()



from typing import TextIO, Optional


class Writer:

    def __init__(
            self,
            out: TextIO,
            pretty_print=True,
            tab_symbol: Optional[str] = None):
        self._out = out
        self._broken = False
        self.pretty_print = pretty_print
        self.tab_symbol = tab_symbol or '  '
        self.indentation = 0

    def write_raw(self, chars: Optional[str]):
        if not chars:
            return

        self._out.write(chars)

    def write(self, chars: Optional[str]):
        if not chars:
            return

        for char in chars:
            if self._broken:
                if self.pretty_print and self.indentation > 0:
                    self._out.write(self.indentation * self.tab_symbol)
                self._broken = False

            self._out.write(char)

            if char == '\n':
                self._broken = True

    def break_line(self):
        if not self._broken:
            self.write('\n')



from stx.compiling.linking.numbering import link_document_numbers
from stx.compiling.linking.referencing import validate_references
from stx.compiling.linking.referencing import auto_generate_special_references
from stx.compiling.parsing.parser import capture, CTX
from stx.functions.core import resolve_document
from stx.compiling.reading.reader import Reader
from stx.document import Document
from stx.utils.thread_context import context


def compile_document(file_path: str) -> Document:
    doc = Document(file_path)
    reader = Reader()
    ctx = CTX(doc, reader)

    context.push_reader(reader)

    reader.push_file(file_path)

    capture(ctx)

    context.pop_reader()

    auto_generate_special_references(doc)

    while True:
        link_document_numbers(doc)

        resolved_components = resolve_document(doc)

        validate_references(doc)

        if resolved_components == 0:
            break

    return doc



from typing import List, Optional

from stx.compiling.reading.content import Content


class Chain:

    def __init__(self, file_paths: List[str]):
        self.file_paths = file_paths
        self._current_content: Optional[Content] = None

        if len(file_paths) == 0:
            raise Exception('no files available')

    def active(self) -> bool:
        content = self.get_current_content()

        return content is not None and not content.halted()

    def load_next_content(self) -> Optional[Content]:
        if len(self.file_paths) == 0:
            return None
        file_path = self.file_paths.pop(0)
        content = Content.from_file(file_path)
        return content

    def get_current_content(self) -> Optional[Content]:
        if self._current_content is None or self._current_content.halted():
            self._current_content = self.load_next_content()
        return self._current_content



from ._captured_text import CapturedText  # noqa: F401
from ._code_block import CodeBlock  # noqa: F401
from ._component import Component, DisplayMode  # noqa: F401
from ._composite import Composite  # noqa: F401
from ._content_box import ContentBox  # noqa: F401
from ._custom_text import CustomText  # noqa: F401
from ._figure import Figure  # noqa: F401
from ._function_call import FunctionCall  # noqa: F401
from ._image import Image  # noqa: F401
from ._layout import Layout  # noqa: F401
from ._link_text import LinkText  # noqa: F401
from ._list_block import ListBlock  # noqa: F401
from ._literal import Literal  # noqa: F401
from ._paragraph import Paragraph  # noqa: F401
from ._plain_text import PlainText  # noqa: F401
from ._section import Section  # noqa: F401
from ._separator import Separator  # noqa: F401
from ._styled_text import StyledText  # noqa: F401
from ._table import Table  # noqa: F401
from ._table_row import TableRow  # noqa: F401
from ._toc import TableOfContents, ElementReference  # noqa: F401



from os import path, walk
from typing import Iterator, List


def resolve_path(base_path: str, relative_path: str) -> str:
    if path.isabs(relative_path):
        return relative_path

    return path.normpath(path.join(base_path, relative_path))


def resolve_sibling(main_path: str, relative_path: str) -> str:
    base_dir = path.dirname(main_path)

    return resolve_path(base_dir, relative_path)


def walk_files(dir_path: str) -> Iterator[str]:
    for root, dirs, files in walk(dir_path):
        for name in files:
            yield path.join(root, name)


def resolve_include_files(include_path: str, source_path: str) -> List[str]:
    target_path = resolve_sibling(source_path, include_path)

    if path.isdir(target_path):
        return [file_path for file_path in sorted(walk_files(target_path))]

    return [target_path]



from typing import Any

from stx.themes.built_in.classic_html import classic_theme

_themes = {}


def register_theme(name: str, output_format: str, theme: Any, override=False):
    key = f'{name}/{output_format}'

    if not override and key in _themes:
        raise Exception(f'Theme is already registered: {name}')

    _themes[key] = theme


def get_theme(name: str, output_format: str) -> Any:
    key = f'{name}/{output_format}'

    if key not in _themes:
        raise Exception(f'Theme `{name}` not found for '
                        f'the `{output_format}` format.')

    return _themes[key]


register_theme('classic', 'html', classic_theme)



from __future__ import annotations

from typing import List, TextIO, Optional

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class ContentBox(Component):  # TODO rename to box

    def __init__(self, location: Location):
        self.location = location
        self.content: Optional[Component] = None
        self.style = None

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.BLOCK

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        self.content.write_text(output)

    def get_children(self) -> List[Component]:
        return [self.content]

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        style_value = attributes.get('style')

        if style_value is not None:
            self.style = style_value.to_str()



from typing import List, Any


def to_str(value: Any) -> str:
    if isinstance(value, str):
        return value
    elif value is None:
        return ''
    return str(value)


def to_list(value: Any) -> list:
    if isinstance(value, list):
        return value
    elif value is None:
        return []
    return [value]


def to_list_str(value: Any) -> List[str]:
    return [to_str(item) for item in to_list(value)]



from __future__ import annotations

import re
from typing import List, Optional, TextIO

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


def is_url(ref: str) -> bool:
    return bool(re.match(r'(?i)^([a-z]+:)?//', ref))


class LinkText(Component):

    def __init__(
            self,
            location: Location,
            contents: List[Component],
            reference: Optional[str]):
        self.location = location
        self.contents = contents
        self.reference = reference
        self.invalid = False

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.INLINE

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        for content in self.contents:
            content.write_text(output)

    def get_children(self) -> List[Component]:
        return self.contents

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass

    def is_internal(self) -> bool:
        return self.reference is not None and not is_url(self.reference)

    def is_external(self) -> bool:
        return self.reference is not None and is_url(self.reference)



from typing import Optional, List

from stx.compiling.reading.location import Location
from stx.components import Component, Composite, CodeBlock, Table, Image, \
    FunctionCall, CustomText
from stx.components import ListBlock, Paragraph, PlainText, StyledText
from stx.components import LinkText, Literal, Figure, Section, Separator
from stx.components import ContentBox, TableOfContents, ElementReference
from stx.components import CapturedText
from stx.document import Document
from stx.utils.stx_error import StxError


def location_to_json(location: Optional[Location]) -> Optional[dict]:
    if location is None:
        return None

    return {
        'file_path': location.file_path,
        'line': location.line,
        'column': location.column,
    }


def extend_base(component: Component, type_id: str, inherited: dict):
    return {
        'type': type_id,
        'refs': component.get_refs(),
        'location': location_to_json(component.location),
        **inherited,
    }


def composite_to_json(composite: Composite) -> dict:
    return extend_base(composite, 'composite', {
        'components': components_to_json(composite.components),
    })


def code_block_to_json(code_block: CodeBlock) -> dict:
    return extend_base(code_block, 'code-block', {
        'lang': code_block.lang,
        'content': components_to_json(code_block.contents),
    })


def table_to_json(table: Table) -> dict:
    return extend_base(table, 'table', {
        'caption': component_to_json(table.caption),
        'number': table.number,
        'rows': [
            {
                'header': row.header,
                'cells': components_to_json(row.cells),
            }
            for row in table.rows
        ]
    })


def list_to_json(list_block: ListBlock) -> dict:
    return extend_base(list_block, 'list', {
        'ordered': list_block.ordered,
        'items': components_to_json(list_block.items),
    })


def paragraph_to_json(paragraph: Paragraph) -> dict:
    return extend_base(paragraph, 'paragraph', {
        'contents': components_to_json(paragraph.contents),
    })


def plain_text_to_json(plain_text: PlainText) -> dict:
    return extend_base(plain_text, 'plain-text', {
        'content': plain_text.content,
    })


def styled_text_to_json(styled_text: StyledText) -> dict:
    return extend_base(styled_text, 'styled-text', {
        'style': styled_text.style,
        'contents': components_to_json(styled_text.contents),
    })


def custom_text_to_json(custom: CustomText) -> dict:
    return extend_base(custom, 'styled-text', {
        'custom_style': custom.custom_style,
        'contents': components_to_json(custom.contents),
    })


def link_text_to_json(link_text: LinkText) -> dict:
    return extend_base(link_text, 'link-text', {
        'reference': link_text.reference,
        'contents': components_to_json(link_text.contents),
    })


def literal_block_to_json(literal: Literal) -> dict:
    return extend_base(literal, 'literal', {
        # TODO add origin?
        'content': literal.text,
    })


def figure_to_json(figure: Figure) -> dict:
    return extend_base(figure, 'figure', {
        'number': figure.number,
        'content': component_to_json(figure.content),
        'caption': component_to_json(figure.caption),
    })


def section_to_json(section: Section) -> dict:
    return extend_base(section, 'section', {
        'level': section.level,
        'number': section.number,
        'heading': component_to_json(section.heading),
        'content': component_to_json(section.content),
    })


def toc_elements_to_json(elements: List[ElementReference]) -> List[dict]:
    return [
        {
            'title': element.title,
            'reference': element.reference,
            'number': element.number,
            'elements': toc_elements_to_json(element.elements),
        }
        for element in elements
    ]


def toc_to_json(toc: TableOfContents) -> dict:
    return extend_base(toc, 'toc', {
        'title': toc.title,
        'elements': toc_elements_to_json(toc.elements),
    })


def separator_to_json(separator: Separator) -> dict:
    return extend_base(separator, 'separator', {
        'name': separator.level,
    })


def box_to_json(box: ContentBox) -> dict:
    return extend_base(box, 'box', {
        'style': box.style,
        'content': component_to_json(box.content),
    })


def components_to_json(components: List[Component]) -> List[dict]:
    return [
        component_to_json(component)
        for component in components
    ]


def captured_text_to_json(captured: CapturedText) -> dict:
    return extend_base(captured, 'captured-text', {
        'class': captured.class_,
        'contents': components_to_json(captured.contents),
    })


def image_to_json(image: Image):
    return extend_base(image, 'image', {
        'src': image.src,
        'alt': image.alt,
    })


def function_call_to_json(call: FunctionCall):
    if call.result is None:
        raise StxError(f'Not resolved function: {call.key}', call.location)

    return component_to_json(call.result)


def component_to_json(content: Optional[Component]) -> Optional[dict]:
    if content is None:
        return None
    elif isinstance(content, Composite):
        return composite_to_json(content)
    elif isinstance(content, CodeBlock):
        return code_block_to_json(content)
    elif isinstance(content, Table):
        return table_to_json(content)
    elif isinstance(content, ListBlock):
        return list_to_json(content)
    elif isinstance(content, Paragraph):
        return paragraph_to_json(content)
    elif isinstance(content, PlainText):
        return plain_text_to_json(content)
    elif isinstance(content, StyledText):
        return styled_text_to_json(content)
    elif isinstance(content, CustomText):
        return custom_text_to_json(content)
    elif isinstance(content, LinkText):
        return link_text_to_json(content)
    elif isinstance(content, Literal):
        return literal_block_to_json(content)
    elif isinstance(content, Figure):
        return figure_to_json(content)
    elif isinstance(content, Section):
        return section_to_json(content)
    elif isinstance(content, TableOfContents):
        return toc_to_json(content)
    elif isinstance(content, Separator):
        return separator_to_json(content)
    elif isinstance(content, ContentBox):
        return box_to_json(content)
    elif isinstance(content, CapturedText):
        return captured_text_to_json(content)
    elif isinstance(content, Image):
        return image_to_json(content)
    elif isinstance(content, FunctionCall):
        return function_call_to_json(content)
    else:
        raise NotImplementedError(f'Not implemented type: {type(content)}')


def document_to_json(doc: Document) -> dict:
    return {
        'title': doc.title,
        'author': doc.author,
        'content': component_to_json(doc.content)
    }



from __future__ import annotations

from typing import List, Dict, Any, Optional


class Value:

    def to_any(self) -> Any:
        raise NotImplementedError()

    def to_map(self) -> Dict[str, Value]:
        raise NotImplementedError()

    def try_token(self) -> Optional[Token]:
        raise NotImplementedError()

    def try_entry(self) -> Optional[Entry]:
        raise NotImplementedError()

    def try_group(self) -> Optional[Group]:
        raise NotImplementedError()

    def try_str(self) -> Optional[str]:
        raise NotImplementedError()

    def try_dict(self) -> Optional[Dict[str, Any]]:
        raise NotImplementedError()

    def try_list(self) -> Optional[List[Any]]:
        raise NotImplementedError()

    def to_token(self) -> Token:
        token = self.try_token()

        if token is None:
            raise Exception('Cannot convert to token.')

        return token

    def to_entry(self) -> Entry:
        entry = self.try_entry()

        if entry is None:
            raise Exception('Cannot convert to entry.')

        return entry

    def to_group(self) -> Group:
        group = self.try_group()

        if group is None:
            raise Exception('Cannot convert to group.')

        return group

    def to_str(self) -> str:
        s = self.try_str()

        if s is None:
            raise Exception('Cannot convert to str.')

        return s

    def to_dict(self) -> Dict[str, Any]:
        d = self.try_dict()

        if d is None:
            raise Exception('Cannot convert to dict.')

        return d

    def to_list(self) -> List[Any]:
        lst = self.try_list()

        if lst is None:
            raise Exception('Cannot convert to list.')

        return lst


class Empty(Value):

    def to_any(self) -> Any:
        return None

    def to_map(self) -> Dict[str, Value]:
        return {}

    def try_token(self) -> Optional[Token]:
        return None

    def try_entry(self) -> Optional[Entry]:
        return None

    def try_group(self) -> Optional[Group]:
        return None

    def try_str(self) -> Optional[str]:
        return None

    def try_dict(self) -> Optional[Dict[str, Any]]:
        return None

    def try_list(self) -> Optional[List[Any]]:
        return None


class Token(Value):

    def __init__(self, text: str):
        self.text = text

    def to_token(self) -> Token:
        return self

    def to_str(self) -> str:
        return self.text

    def to_any(self) -> Any:
        return self.text

    def to_map(self) -> Dict[str, Value]:
        return {self.text: Empty()}

    def try_token(self) -> Optional[Token]:
        return self

    def try_entry(self) -> Optional[Entry]:
        return Entry(self.text, Empty())

    def try_group(self) -> Optional[Group]:
        return Group([self])

    def try_str(self) -> Optional[str]:
        return self.text

    def try_dict(self) -> Optional[Dict[str, Any]]:
        return {self.text: None}

    def try_list(self) -> Optional[List[Any]]:
        return [self.text]


class Entry(Value):

    def __init__(self, name: str, value: Value):
        self.name = name
        self.value = value

    def to_entry(self) -> Entry:
        return self

    def to_dict(self):
        return {self.name: self.value.to_any()}

    def to_any(self) -> Any:
        return self.to_dict()

    def to_map(self) -> Dict[str, Value]:
        return {self.name: self.value}

    def try_token(self) -> Optional[Token]:
        if isinstance(self.value, Empty):
            return Token(self.name)
        return None

    def try_entry(self) -> Optional[Entry]:
        return self

    def try_group(self) -> Optional[Group]:
        return Group([self])

    def try_str(self) -> Optional[str]:
        if isinstance(self.value, Empty):
            return self.name
        return None

    def try_dict(self) -> Optional[Dict[str, Any]]:
        return self.to_dict()

    def try_list(self) -> Optional[List[Any]]:
        return [self.to_any()]


class Group(Value):

    def __init__(self, items: List[Value]):
        self.items = items

    def to_group(self) -> Group:
        return self

    def try_group(self) -> Optional[Group]:
        return self

    def to_list(self) -> List[Any]:
        return [item.to_any() for item in self.items]

    def try_list(self) -> Optional[List[Any]]:
        return self.to_list()

    def to_any(self) -> Any:
        entries = 0
        tokens = 0
        other = 0

        for item in self.items:
            if isinstance(item, Entry):
                entries += 1
            elif isinstance(item, Token):
                tokens += 1
            else:
                other += 1

        if entries > 0 and other == 0:
            d = self.try_dict()
            if d is not None:
                return d

        return self.to_list()

    def to_map(self) -> Dict[str, Value]:
        d = {}

        for item in self.items:
            entry = item.to_entry()

            d[entry.name] = entry.value

        return d

    def try_token(self) -> Optional[Token]:
        length = len(self.items)

        if length == 1:
            return self.items[0].try_token()

        return None

    def try_entry(self) -> Optional[Entry]:
        length = len(self.items)

        if length == 1:
            key = self.items[0].try_str()
            value = None
        elif length == 2:
            key = self.items[0].try_str()
            value = self.items[1]
        else:
            key = None
            value = None

        if key is not None:
            if value is None:
                value = Empty()
            return Entry(key, value)
        return None

    def try_str(self) -> Optional[str]:
        length = len(self.items)

        if length == 1:
            return self.items[0].try_str()

        return None

    def try_dict(self) -> Optional[Dict[str, Any]]:
        d = {}

        for item in self.items:
            entry = item.try_entry()

            if entry is None or entry.name in d:
                return None

            d[entry.name] = entry.value.to_any()

        return d



from typing import List, Optional

from stx.compiling.reading.chain import Chain
from stx.compiling.reading.content import Content
from stx.compiling.reading.location import Location


class Reader:

    def __init__(self):
        self._chain_stack: List[Chain] = []

    def push_file(self, file_path: str):
        self.push_files([file_path])

    def push_files(self, file_paths: List[str]):
        self._chain_stack.append(Chain(file_paths))

    def active(self) -> bool:
        chain = self.get_chain()

        return chain is not None and chain.active()

    def get_chain(self) -> Optional[Chain]:
        while True:
            if len(self._chain_stack) == 0:
                return None

            chain = self._chain_stack[-1]

            if chain.active():
                return chain

            self._chain_stack.pop()

    def get_content(self) -> Content:
        chain = self.get_chain()
        content = chain.get_current_content()

        if content is None:
            if len(self._chain_stack) == 0:
                raise Exception('no more chains!')

            self._chain_stack.pop()

            chain = self.get_chain()
            content = chain.get_current_content()

        if content is None:
            raise Exception('no more content')

        return content

    def get_location(self) -> Optional[Location]:
        if not self.active():
            return None

        return self.get_content().get_location()



from typing import TextIO, Any

entities = {
    # Main Entities
    160: '&nbsp;',
    60: '&lt;',
    62: '&gt;',
    38: '&amp;',
    34: '&quot;',
    39: '&apos;',

    # Mathematical Symbols
    8704: '&forall;',
    8706: '&part;',
    8707: '&exist;',
    8709: '&empty;',
    8711: '&nabla;',
    8712: '&isin;',
    8713: '&notin;',
    8715: '&ni;',
    8719: '&prod;',
    8721: '&sum;',

    # Greek Letters
    913: '&Alpha;',
    914: '&Beta;',
    915: '&Gamma;',
    916: '&Delta;',
    917: '&Epsilon;',
    918: '&Zeta;',

    # Other Entities
    162: '&cent;',
    163: '&pound;',
    165: '&yen;',
    169: '&copy;',
    174: '&reg;',
    8364: '&euro;',
    8482: '&trade;',
    8592: '&larr;',
    8593: '&uarr;',
    8594: '&rarr;',
    8595: '&darr;',
    9824: '&spades;',
    9827: '&clubs;',
    9829: '&hearts;',
    9830: '&diams;',
}


def write_text(value: Any, out: TextIO):
    if value is None:
        return

    for c in str(value):
        code = ord(c)
        entity = entities.get(code)

        if entity is not None:
            out.write(entity)
        # Non-text control codes
        elif code not in [9, 10, 13] and (code <= 31 or (127 <= code <= 159)):
            out.write(f'&#{code};')
        else:
            out.write(c)



from __future__ import annotations

from typing import List, Optional, TextIO

from ._component import Component, DisplayMode
from ._table_row import TableRow
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class Table(Component):

    def __init__(self, location: Location):
        self.location = location
        self.rows: List[TableRow] = []
        self.caption: Optional[Component] = None
        self.number: Optional[str] = None

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.BLOCK

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        for row in self.rows:
            row.write_text(output)

    def get_children(self) -> List[Component]:
        if self.caption is not None:
            return [self.caption, *self.rows]

        return self.rows

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass

    def get_last_row(self) -> Optional[TableRow]:
        if len(self.rows) > 0:
            return self.rows[-1]
        return None



from stx.functions import utils
from stx.components import Component, FunctionCall, PlainText
from stx.document import Document


def resolve_line_feed(document: Document, call: FunctionCall) -> Component:
    utils.check_unknown_options(call.options, call)

    # TODO create component to represent line feed
    return PlainText(call.location, '\n')



from __future__ import annotations

import re
from io import StringIO
from typing import List, Optional

from stx import logger
from stx.compiling.marks import get_matching_mark, mark_max_length
from stx.compiling.reading.location import Location
from stx.utils.stx_error import StxError
from stx.utils.debug import see

LF_CHAR = '\n'  # Line Feed

EMPTY_OR_WHITESPACE = r'^ *$'


class TRX:

    def __init__(self, content: Content):
        self._content = content
        self._position = None
        self._line = None
        self._column = None
        self._state = 'created'

    def __enter__(self):
        if self._state == 'started':
            raise Exception('Transaction is already started.')

        self._position = self._content.position
        self._line = self._content.line
        self._column = self._content.column
        self._state = 'started'
        self._content.transactions.append(self)
        return self

    def _pop_transaction(self):
        if len(self._content.transactions) == 0:
            raise Exception('Corrupted transaction')

        trx = self._content.transactions.pop()

        if trx != self:
            raise Exception('Corrupted transaction')

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            raise exc_val

        if self._state not in ['saved', 'canceled']:
            raise Exception(f'Unfinished transaction: {self._state}')

    def save(self):
        self._pop_transaction()

        self._state = 'saved'

    def cancel(self):
        self._pop_transaction()

        self._content.position = self._position
        self._content.line = self._line
        self._content.column = self._column

        self._state = 'canceled'


class Content:

    def __init__(self, content: str, file_path: str):
        self.file_path = file_path
        self.position = 0
        self.line = 0
        self.column = 0
        self.transactions: List[TRX] = []
        self._content = content
        self._length = len(content)

    @staticmethod
    def from_file(file_path: str) -> Content:
        logger.info(f'Loading file {see(file_path, None)}...')

        with open(file_path, mode='r') as stream:
            content = stream.read().rstrip()

        return Content(content, file_path)

    def checkout(self) -> TRX:
        return TRX(self)

    def go_back(self, location: Location):
        self.position = location.position
        self.line = location.line
        self.column = location.column

    def halted(self):
        return self.position >= self._length and len(self.transactions) == 0

    def move_next(self):
        if self.position < self._length:
            new_line = (self._content[self.position] == LF_CHAR)

            self.position += 1

            if new_line:
                self.line += 1
                self.column = 0
            else:
                self.column += 1
        else:
            raise StxError('EOF')

    def peek(self) -> Optional[str]:
        if self.position < self._length:
            return self._content[self.position]

        return None

    def test_any(self, tokens: List[str]) -> bool:
        for token in tokens:
            if self.test(token):
                return True

        return False

    def test(self, token: Optional[str]) -> bool:
        if token is None:
            return False

        loc0 = self.get_location()

        for expected_char in token:
            actual_char = self.peek()

            if actual_char is None or actual_char != expected_char:
                self.go_back(loc0)
                return False

            self.move_next()

        self.go_back(loc0)
        return True

    def pull_any(self, tokens: List[str]) -> Optional[str]:
        for token in tokens:
            if self.pull(token):
                return token
        return None

    def pull(self, token: str) -> bool:
        if token is None:
            return False

        loc0 = self.get_location()

        for expected_char in token:
            actual_char = self.peek()

            if actual_char is None or actual_char != expected_char:
                self.go_back(loc0)
                return False

            self.move_next()

        return True

    def read_next(self) -> Optional[str]:
        c = self.peek()

        if c is None:
            return None

        self.move_next()
        return c

    def read_until(
            self,
            chars: List[str],
            consume_last=False,
            max_length=None) -> str:
        out = StringIO()

        while True:
            if max_length is not None and out.tell() >= max_length:
                break

            c = self.peek()

            if c is None:
                break
            elif c in chars:
                if consume_last:
                    out.write(c)
                    self.move_next()
                break

            out.write(c)
            self.move_next()

        return out.getvalue()

    def read_while(self, chars: List[str]) -> str:
        out = StringIO()

        while True:
            c = self.peek()

            if c is None or c not in chars:
                break

            out.write(c)

            self.move_next()

        return out.getvalue()

    def read_max(self, max_length: int) -> str:
        out = StringIO()

        while True:
            c = self.peek()

            if c is None:
                break

            out.write(c)

            self.move_next()

            if out.tell() >= max_length:
                break

        return out.getvalue()

    def alive(self, indentation: int) -> bool:  # TRX
        if indentation > 0 and self.column < indentation:
            with self.checkout() as trx:
                line_prefix = self.read_until(['\n'], max_length=indentation)

                if re.match(EMPTY_OR_WHITESPACE, line_prefix):
                    trx.save()
                    return True
                else:
                    trx.cancel()
                    return False

        return self.position < self._length

    def read_mark(self, valid_marks: List[str]) -> Optional[str]:
        loc0 = self.get_location()

        token = self.read_until(
            chars=[' ', '\n'],
            consume_last=False,
            max_length=mark_max_length)

        mark = get_matching_mark(token, valid_marks)

        self.go_back(loc0)

        if mark is None:
            return None

        for i in range(len(mark)):
            self.move_next()

        return mark

    def read_line(self, indentation: int) -> Optional[str]:
        loc0 = self.get_location()

        while self.column < indentation:
            if self.peek() != ' ':
                self.go_back(loc0)
                return None

            self.move_next()

        out = StringIO()
        length = 0

        while True:
            c = self.peek()

            if c is None:
                if length == 0:
                    # Nothing has been actually read
                    self.go_back(loc0)
                    return None
                else:
                    # Not a line feed but it's ok
                    break

            out.write(c)
            self.move_next()
            length += 1

            if c == '\n':
                # Line feed: end of line!
                break

        return out.getvalue()

    def expect_char(self, options: List[str]):
        c = self.read_next()

        if c not in options:
            raise StxError(f'Expected any of {options}')

        return c

    def expect_end_of_line(self):
        with self.checkout() as trx:
            text = self.read_until(['\n'], consume_last=True)

            if len(text.strip()) > 0:
                trx.cancel()
                raise StxError('Expected end of line.')

            trx.save()

    def get_location(self) -> Location:
        return Location(self.file_path, self.line, self.column, self.position)

    def read_spaces(self, max_length: int = None) -> int:
        count = 0

        while self.peek() == ' ':
            self.move_next()

            count += 1

            if max_length is not None and count >= max_length:
                break

        return count

    def consume_indentation(self, indentation: int) -> bool:
        loc0 = self.get_location()

        while self.column < indentation:
            if self.peek() == ' ':
                self.move_next()
            else:
                break

        if self.column >= indentation:
            return True

        self.go_back(loc0)
        return False

    def consume_empty_line(self) -> bool:
        loc0 = self.get_location()

        count = 0

        while self.peek() == ' ':
            self.move_next()

            count += 1

        c = self.peek()

        if c is None:
            return count > 0
        elif c == '\n':
            self.move_next()
            return True

        self.go_back(loc0)
        return False



from typing import Optional

from stx.functions import utils
from stx.components import FunctionCall, Component, ContentBox
from stx.document import Document


def resolve_admonition(document: Document, call: FunctionCall) -> Component:
    options = utils.make_options_dict(call, key_for_str='type')

    data_type = options.pop('type', None)

    utils.check_unknown_options(options, call)

    return make_admonition(document, call, data_type)


def resolve_warning(document: Document, call: FunctionCall) -> Component:
    utils.check_unknown_options(call.options, call)

    return make_admonition(document, call, 'warning')


def resolve_information(document: Document, call: FunctionCall) -> Component:
    utils.check_unknown_options(call.options, call)

    return make_admonition(document, call, 'information')


def make_admonition(
        document: Document,
        call: FunctionCall,
        data_type: Optional[str]) -> Component:
    if call.argument is None:
        raise call.error('Admonitions require a captured component.')

    box = ContentBox(call.location)
    box.content = call.argument
    box.style = data_type

    return box



from __future__ import annotations

from enum import Enum
from io import StringIO
from typing import List, Iterable, TextIO, Union, Optional

from stx.compiling.reading.location import Location
from stx.data_notation.values import Value
from stx.utils.stx_error import StxError
from stx.utils.tracked_dict import TrackedDict


class DisplayMode(Enum):
    DEFAULT = 'default'
    INLINE = 'inline'
    BLOCK = 'block'

    @staticmethod
    def compute_display_mode(components: List[Component]) -> DisplayMode:
        inlines = 0
        blocks = 0
        other = 0

        for component in components:
            if component.display_mode == DisplayMode.BLOCK:
                blocks += 1
            elif component.display_mode == DisplayMode.INLINE:
                inlines += 1
            else:
                other += 1

        if inlines > 0 and blocks == 0:
            return DisplayMode.INLINE
        elif blocks > 0 and inlines == 0:
            return DisplayMode.BLOCK

        return DisplayMode.DEFAULT


class Component:
    ref: Union[str, List[str], None] = None
    location: Optional[Location] = None

    def get_refs(self) -> List[str]:
        if self.ref is None:
            return []
        elif isinstance(self.ref, list):
            return [str(r) for r in self.ref]
        return [str(self.ref)]

    def get_main_ref(self) -> Optional[str]:
        refs = self.get_refs()

        if len(refs) == 0:
            return None

        return refs[0]

    def get_other_refs(self) -> List[str]:
        refs = self.get_refs()

        if len(refs) > 0:
            refs.pop(0)

        return refs

    def add_ref(self, ref):
        refs = self.get_refs()
        if ref not in refs:
            refs.append(ref)
        self.ref = refs

    def has_refs(self) -> bool:
        if isinstance(self.ref, list):
            return len(self.ref) > 0
        return self.ref is not None

    def is_rich(self) -> bool:
        raise NotImplementedError()

    def get_text(self) -> str:
        output = StringIO()

        self.write_text(output)

        return output.getvalue()

    def walk(self) -> Iterable[Component]:
        for child in self.get_children():
            yield child

            for c in child.walk():
                yield c

    def write_text(self, output: TextIO):
        raise NotImplementedError()

    def get_children(self) -> List[Component]:
        raise NotImplementedError()

    @property
    def display_mode(self) -> DisplayMode:
        raise NotImplementedError()

    def apply_attributes(self, attributes: TrackedDict[str, Value]):
        ref_value = attributes.get('ref')

        if ref_value is not None:
            ref_token = ref_value.try_token()

            if ref_token is not None:
                self.ref = ref_token.to_str()
            else:
                self.ref = ref_value.to_list()

        self.apply_advanced_attributes(attributes)

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        raise NotImplementedError()

    def error(self, message: str) -> Exception:
        return StxError(message, self.location)



from typing import TextIO

from stx.outputs.html.serializer import document_to_html
from stx.outputs.html.themes import HtmlTheme, NullHtmlTheme
from stx.outputs.output_action import OutputAction
from stx.themes.registry import get_theme


class HtmlOutputAction(OutputAction):

    def dump(self, out: TextIO):
        if self.theme is not None:
            theme = get_theme(self.theme, self.format_key)

            if not isinstance(theme, HtmlTheme):
                raise Exception(f'Theme `{self.theme}` is not '
                                f'compatible with `{self.format_key}`.')
        else:
            theme = NullHtmlTheme()

        html = document_to_html(self.document, theme)

        for tag in html:
            tag.render(out)

        # TODO implement pretty-print



from typing import List, Optional

from stx.components import Component, Figure, Table
from stx.data_notation.values import Value
from stx.utils.stx_error import StxError
from stx.utils.debug import see
from stx.utils.tracked_dict import TrackedDict


class Composer:

    def __init__(self):
        self.stack: List[List[Component]] = []
        self.attributes_buffer: TrackedDict[str, Value] = TrackedDict({})
        self.pre_captions: List[Component] = []

    def push(self):
        self.stack.append([])

    def pop(self) -> Optional[List[Component]]:
        return self.stack.pop()

    @property
    def components(self) -> List[Component]:
        if len(self.stack) == 0:
            raise StxError('Empty stack.')

        return self.stack[-1]

    def add(self, component: Component):
        if not self.attributes_buffer.empty():
            component.apply_attributes(self.attributes_buffer)

            unknown_attr_keys = self.attributes_buffer.unknown_keys()

            if len(unknown_attr_keys) > 0:
                # TODO improve error message
                raise StxError(f'Unknown attributes: {unknown_attr_keys} '
                               f'for {type(component)}.')

            self.attributes_buffer.clear()

        if len(self.pre_captions) > 0:
            if isinstance(component, Table):
                component.caption = self.pre_captions.pop()

                self.components.append(component)
            else:
                figure = Figure(
                    component.location, component, self.pre_captions.pop())

                self.components.append(figure)
        else:
            self.components.append(component)

    def push_attribute(self, key: str, value: Value):
        if key in self.attributes_buffer.keys():
            raise StxError(f'The attribute {see(key)} was already defined.')

        self.attributes_buffer.put(key, value)

    def get_last_component(self) -> Optional[Component]:
        comps = self.components

        if len(comps) == 0:
            return None

        return comps[-1]

    def push_pre_caption(self, caption: Component):
        self.pre_captions.append(caption)

    def push_post_caption(self, caption: Component):
        consumed = False

        for comps in reversed(self.stack):
            if len(comps) > 0:
                component = comps[-1]

                if isinstance(component, Table):
                    component.caption = caption
                else:
                    figure = Figure(component.location, component, caption)

                    comps[-1] = figure

                consumed = True
                break

        if not consumed:
            raise StxError('Expected component for post-caption.')



import os
import threading
import time
import traceback

import click
import pkg_resources
from click import UsageError

from watchdog.events import FileSystemEventHandler, FileSystemEvent
from watchdog.observers import Observer

from stx import logger
from stx.compiling.compiler import compile_document
from stx.document import Document
from stx.outputs import OutputFile, OutputAction
from stx.utils.debug import see

app_name = 'stx'
app_version = pkg_resources.require(app_name)[0].version
app_title = f'{app_name} {app_version}'


def process_file(input_file: str) -> Document:
    logger.info(f'Processing file {see(input_file)}...')

    document = compile_document(input_file)

    if len(document.actions) == 0:
        logger.warning('No actions were registered.')
    else:
        for action in document.actions:
            action.run()

    return document


def watch_file(input_file: str, debug_mode: bool):
    ignored_files = set()

    def refresh():
        try:
            document = process_file(input_file)

            ignored_files.clear()

            for action in document.actions:
                if (isinstance(action, OutputAction)
                        and isinstance(action.target, OutputFile)):
                    ignored_files.add(
                        os.path.abspath(action.target.file_path)
                    )
        except Exception as e:
            print(f'Error: {e}')
            if debug_mode:
                print(traceback.format_exc())

    class Handler(FileSystemEventHandler):

        def __init__(self):
            self.lock = threading.Lock()

        def on_any_event(self, event: FileSystemEvent):
            if event.is_directory:
                return

            with self.lock:
                src_file = os.path.abspath(event.src_path)

                if src_file not in ignored_files:
                    logger.info(f'Detected change on: {src_file}')

                    refresh()

    refresh()

    root_dir = os.path.dirname(input_file)

    logger.info(f'Watching directory {see(root_dir)} for changes...')

    observer = Observer()
    observer.schedule(Handler(), root_dir, recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        logger.info('Stopping watch...')
        observer.stop()
    observer.join()


def main(input_file: str, watch_mode=False, version=False, debug_mode=False):
    if version:
        print(app_title)
    elif input_file == '':
        raise UsageError('Missing input file.')

    input_file = os.path.abspath(input_file)

    if watch_mode:
        watch_file(input_file, debug_mode)
    else:
        process_file(input_file)


@click.command(name='stx')
@click.argument('input_file', default='')
@click.option(
    '-w', '--watch', help='Watches the document for changes.',
    is_flag=True, default=False)
@click.option(
    '-v', '--version', help='Shows the STX version.',
    is_flag=True, default=False)
@click.option(
    '-d', '--debug', help='Enables debug mode.',
    is_flag=True, default=False)
def cli(input_file: str, watch: bool, version: bool, debug):
    """Processes the STX document indicated by INPUT_FILE."""
    main(input_file, watch, version, debug)



from __future__ import annotations

from typing import List, TextIO

from ._component import Component, DisplayMode

# TODO looks like this is not a component per se

from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class TableRow(Component):

    def __init__(self, location: Location, header: bool):
        self.location = location
        self.header = header
        self.cells: List[Component] = []

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.DEFAULT

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        for cell in self.cells:
            cell.write_text(output)

    def get_children(self) -> List[Component]:
        return self.cells

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



import sys

from stx.compiling.reading.location import Location


def debug(message: str, location: Location = None):
    print(message if location is None else location.decorate(message))


def info(message: str, location: Location = None):
    print(message if location is None else location.decorate(message))


def warning(message: str, location: Location = None):
    print(
        message if location is None else location.decorate(message),
        file=sys.stderr)



from __future__ import annotations

from typing import List, TextIO, Optional

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class FunctionCall(Component):

    def __init__(
            self,
            location: Location,
            inline: bool,
            key: str,
            options: Value,
            argument: Optional[Component] = None):
        self.location = location
        self.inline = inline
        self.key = key
        self.options = options
        self.argument = argument
        self.result: Optional[Component] = None

    @property
    def display_mode(self) -> DisplayMode:
        if self.result is not None:
            return self.result.display_mode
        return DisplayMode.DEFAULT

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        if self.result is not None:
            self.result.write_text(output)

    def get_children(self) -> List[Component]:
        if self.result is not None:
            return [self.result]
        return []

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from typing import Optional

from stx.compiling.reading.location import Location
from stx.utils.thread_context import context


def generate_error_message(message: str, location: Optional[Location]):
    if location is None:
        reader = context.parser

        if reader is not None:
            content = reader.get_content()
            location = content.get_location()

    if location is None:
        return message

    return location.decorate(message)


class StxError(Exception):

    # TODO make location required
    def __init__(self, message: str, location: Location = None):
        super().__init__(generate_error_message(message, location))





class Location:

    def __init__(self, file_path: str, line: int, column: int, position: int):
        self.file_path = file_path
        self.line = line
        self.column = column
        self.position = position

    def __str__(self):
        return (f'{self.file_path} @'
                f' Line {self.line + 1},'
                f' Column {self.column + 1}')

    def decorate(self, message: str):
        return f'{message} << {self}'



from typing import List

from gramat.errors import GramatError
from gramat.lexing.nodes import SyntaxNode, ContainerNode, TokenNode

from stx.compiling.reading.location import Location
from stx.functions import utils
from stx.components import FunctionCall, Component, CodeBlock, \
    PlainText, CustomText
from stx.document import Document
from stx.grammars import registry


def generate_component(
        location: Location,
        nodes: List[SyntaxNode],
        components: List[Component]):
    for node in nodes:
        if node.rule is not None:
            wrapper = CustomText(location, [], node.rule)

            target = wrapper.contents

            components.append(wrapper)
        else:
            target = components

        if isinstance(node, TokenNode):
            target.append(PlainText(location, node.token))
        elif isinstance(node, ContainerNode):
            generate_component(location, node.nodes, target)
        else:
            raise GramatError(f'unknown node: {node}')


def resolve_code(document: Document, call: FunctionCall) -> Component:
    if call.argument is None:
        raise call.error('Code function requires a captured component.')

    options = utils.make_options_dict(call, key_for_str='lang')

    lang = options.pop('lang', None)

    utils.check_unknown_options(options, call)

    contents = None

    if lang is not None:
        grammar = registry.get_grammar(lang)

        if grammar is not None:
            text = utils.make_plain_text(call.argument)

            try:
                contents = []
                nodes = grammar.tokenize(text)
                generate_component(call.location, nodes, contents)
            except GramatError:
                contents = None

    if contents is None:
        contents = utils.make_component_list(call.argument)

    return CodeBlock(call.location, contents, lang=lang)



from stx.functions import utils
from stx.components import FunctionCall
from typing import List

from stx.components import Component, TableOfContents, ElementReference, \
    Section, Composite
from stx.document import Document


# TODO implement index of tables and figures


def resolve_toc(document: Document, call: FunctionCall) -> Component:
    options = utils.make_options_dict(call, key_for_str='title')

    title = options.pop('title', None)

    utils.check_unknown_options(options, call)

    toc = TableOfContents(call.location, title)

    generate_toc(document, toc)

    return toc


def generate_toc(document: Document, toc: TableOfContents):
    if document.content is not None:
        collect_section_toc(document.content, toc.elements)


def collect_section_toc(
        component: Component, elements: List[ElementReference]):
    if isinstance(component, Section):
        element = ElementReference(
            title=component.heading.get_text().strip(),
            reference=component.get_main_ref(),
            number=component.number,
        )

        elements.append(element)

        collect_section_toc(component.content, element.elements)
    elif isinstance(component, Composite):
        for child in component.components:
            collect_section_toc(child, elements)



from __future__ import annotations

from typing import List, TextIO, Optional

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class Section(Component):

    def __init__(
            self,
            location: Location,
            level: int):
        self.location = location
        self.level = level
        self.heading: Optional[Component] = None
        self.content: Optional[Component] = None
        self.type: Optional[str] = None
        self.number: Optional[str] = None

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.BLOCK

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        self.heading.write_text(output)
        self.content.write_text(output)

    def get_children(self) -> List[Component]:
        return [self.heading, self.content]

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from __future__ import annotations

from typing import List, Optional, TextIO

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class ElementReference:

    def __init__(self, title: str, reference: str, number: Optional[str]):
        self.title = title
        self.reference = reference
        self.number = number
        self.elements: List[ElementReference] = []


class TableOfContents(Component):

    def __init__(self, location: Location, title: Optional[str]):
        self.location = location
        self.title = title
        self.elements: List[ElementReference] = []

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.BLOCK

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        pass

    def get_children(self) -> List[Component]:
        return []

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from ._admonition import resolve_admonition  # noqa: F401
from ._admonition import resolve_warning  # noqa: F401
from ._admonition import resolve_information  # noqa: F401
from ._code import resolve_code  # noqa: F401
from ._custom_style import resolve_custom_style  # noqa: F401
from ._embed import resolve_embed  # noqa: F401
from ._layout import layout_function  # noqa: F401
from ._image import resolve_image  # noqa: F401
from ._line_feed import resolve_line_feed  # noqa: F401
from ._toc import resolve_toc  # noqa: F401



from typing import List

from stx import app, logger
from stx.components import (
    Component, Composite, CodeBlock, Table, Image,
    FunctionCall, CustomText, Layout)
from stx.components import ListBlock, Paragraph, PlainText, StyledText
from stx.components import LinkText, Literal, Figure, Section, Separator
from stx.components import ContentBox, TableOfContents, ElementReference
from stx.components import CapturedText
from stx.document import Document
from stx.outputs.html.dom import Tag
from stx.outputs.html.themes import HtmlTheme
from stx.utils.stx_error import StxError

TYPE_H_TAGS = {
    'chapter': 'h1',
    'sect1': 'h1',
    'sect2': 'h2',
    'sect3': 'h3',
    'sect4': 'h4',
    'sect5': 'h5',
}

CAPTURED_CLASS_TAGS = [
    'cite', 'ins', 'kbd', 'mark', 'q', 's',
    'samp', 'small', 'sub', 'sup', 'var',
]


def document_to_html(document: Document, theme: HtmlTheme) -> List[Tag]:
    logger.info('Serializing HTML5 book...')

    html = Tag('html')

    generate_head(document, html, theme)

    generate_body(document, html, theme)

    return [Tag('!DOCTYPE html'), html]


def generate_head(document: Document, html: Tag, theme: HtmlTheme):
    head = html.append_tag('head')

    if document.encoding:
        head.append_tag('meta', {'charset': document.encoding})

    if document.author:
        head.append_tag('meta', {
            'name': 'author',
            'content': document.author,
        })

    head.append_tag('meta', {
        'name': 'generator',
        'content': app.app_title,
    })

    if document.title:
        head.append_tag('title', text=document.title)

    for css in theme.head_styles:
        head.append_tag('style', {
            'type': 'text/css',
        }, text=css, text_literal=True)

    for stylesheet in document.stylesheets:
        head.append_tag('link', {
            'rel': 'stylesheet',
            'type': 'text/css',
            'href': stylesheet,
        })

    for js in theme.head_scripts:
        head.append_tag('script', {
            'type': 'application/javascript',
        }, text=js, text_literal=True)


def generate_body(document: Document, html: Tag, theme: HtmlTheme):
    body = html.append_tag('body', {'data-type': 'book'})

    if document.title is not None:
        body.append_tag('h1', text=document.title)

    if document.header is not None:
        header = body.append_tag('header')

        generate_component(header, document.header)

    if document.content is not None:
        generate_component(body, document.content)

    if document.footer is not None:
        footer = body.append_tag('footer')

        generate_component(footer, document.footer)

    for js in theme.body_scripts:
        body.append_tag('script', {
            'type': 'application/javascript',
        }, text=js, text_literal=True)


def generate_component(
        parent: Tag, component: Component,
        collapse_paragraph=False,
        collapse_section=False):
    if isinstance(component, Composite):
        generate_components(parent, component.components)
    elif isinstance(component, CodeBlock):
        generate_code_block(parent, component)
    elif isinstance(component, Table):
        generate_table(parent, component)
    elif isinstance(component, ListBlock):
        generate_list_block(parent, component)
    elif isinstance(component, Paragraph):
        generate_paragraph(parent, component, collapse_paragraph)
    elif isinstance(component, PlainText):
        generate_plain_text(parent, component)
    elif isinstance(component, StyledText):
        generate_styled_text(parent, component)
    elif isinstance(component, CustomText):
        generate_custom_text(parent, component)
    elif isinstance(component, LinkText):
        generate_link_text(parent, component)
    elif isinstance(component, CapturedText):
        generate_captured_text(parent, component)
    elif isinstance(component, Literal):
        generate_literal(parent, component)
    elif isinstance(component, Figure):
        generate_figure(parent, component)
    elif isinstance(component, Section):
        generate_section(parent, component, collapse_section)
    elif isinstance(component, TableOfContents):
        generate_toc(parent, component)
    elif isinstance(component, Separator):
        generate_separator(parent, component)
    elif isinstance(component, ContentBox):
        generate_box(parent, component)
    elif isinstance(component, Image):
        generate_image(parent, component)
    elif isinstance(component, FunctionCall):
        generate_function_call(parent, component)
    elif isinstance(component, Layout):
        generate_layout(parent, component)
    else:
        raise NotImplementedError(f'Not implemented type: {type(component)}')


def generate_components(parent: Tag, components: List[Component]):
    for child in components:
        generate_component(parent, child)


def generate_code_block(parent: Tag, code_block: CodeBlock):
    pre = append_component_tag(parent, code_block, 'pre')

    pre.preserve_spaces = True

    pre['data-type'] = 'programlisting'

    if code_block.lang is not None:
        pre['data-code-language'] = code_block.lang

    generate_components(pre, code_block.contents)


def generate_table(parent: Tag, table: Table):
    table_tag = append_component_tag(parent, table, 'table')

    if table.caption is not None:
        caption_tag = table_tag.append_tag('caption')

        if table.number is not None:
            caption_tag['data-number'] = table.number

        generate_component(
            caption_tag, table.caption, collapse_paragraph=True)

    for row in table.rows:
        row_tag = table_tag.append_tag('tr')

        if row.header:
            cell_tag_name = 'th'
        else:
            cell_tag_name = 'td'

        for cell in row.cells:
            cell_tag = row_tag.append_tag(cell_tag_name)

            generate_component(cell_tag, cell, collapse_paragraph=True)


def generate_list_block(parent: Tag, list_block: ListBlock):
    if list_block.ordered:
        list_tag_name = 'ol'
    else:
        list_tag_name = 'ul'

    list_tag = append_component_tag(parent, list_block, list_tag_name)

    for item in list_block.items:
        item_tag = list_tag.append_tag('li')

        generate_component(item_tag, item, collapse_paragraph=True)


def generate_paragraph(parent: Tag, paragraph: Paragraph, collapse: bool):
    if not collapse:
        p_tag = append_component_tag(parent, paragraph, 'p')
    else:
        p_tag = parent

    generate_components(p_tag, paragraph.contents)


def generate_plain_text(parent: Tag, plain_text: PlainText):
    for ref in plain_text.get_refs():
        parent.append_tag('a', {'id': ref}, text='')

    parent.append_text(plain_text.content)


def generate_styled_text(parent: Tag, styled_text: StyledText):
    if styled_text.style == 'strong':
        tag_name = 'strong'
    elif styled_text.style == 'emphasized':
        tag_name = 'em'
    elif styled_text.style == 'code':
        tag_name = 'code'
    elif styled_text.style == 'deleted':
        tag_name = 'del'
    else:
        tag_name = None

    if tag_name is not None:
        styled_tag = append_component_tag(parent, styled_text, tag_name)

        generate_components(styled_tag, styled_text.contents)
    else:
        if styled_text.style == 'double-quote':
            parent.append_literal('&ldquo;')
            generate_components(parent, styled_text.contents)
            parent.append_literal('&rdquo;')
        elif styled_text.style == 'single-quote':
            parent.append_literal('&lsquo;')
            generate_components(parent, styled_text.contents)
            parent.append_literal('&rsquo;')
        else:
            raise StxError(
                f'Not supported style: {styled_text.style}.',
                styled_text.location)


def generate_custom_text(parent: Tag, styled_text: CustomText):
    styled_tag = append_component_tag(parent, styled_text, 'span')

    if styled_text.custom_style is not None:
        styled_tag['class'] = styled_text.custom_style

    generate_components(styled_tag, styled_text.contents)


def generate_link_text(parent: Tag, link_text: LinkText):
    a_tag = append_component_tag(parent, link_text, 'a')

    if link_text.is_internal():
        a_tag['href'] = f'#{link_text.reference}'
        a_tag['data-type'] = 'xref'
    elif link_text.is_external():
        a_tag['href'] = link_text.reference

    if link_text.invalid:
        a_tag['data-status'] = 'invalid'

    generate_components(a_tag, link_text.contents)


def generate_captured_text(parent: Tag, captured: CapturedText):
    if captured.class_ in CAPTURED_CLASS_TAGS:
        cap_tag = append_component_tag(parent, captured, captured.class_)
    else:
        cap_tag = append_component_tag(parent, captured, 'span')

        if captured.class_:
            cap_tag['class'] = captured.class_

    generate_components(cap_tag, captured.contents)


def generate_literal(parent: Tag, literal: Literal):
    parent.append_literal(literal.text)


def generate_figure(parent: Tag, figure: Figure):
    figure_tag = append_component_tag(parent, figure, 'figure')

    generate_component(figure_tag, figure.content)

    caption_tag = figure_tag.append_tag('figcaption')

    if figure.number is not None:
        caption_tag['data-number'] = figure.number

    generate_component(caption_tag, figure.caption, collapse_paragraph=True)


def generate_section(parent: Tag, section: Section, collapse: bool):
    if not collapse:
        section_tag = append_component_tag(parent, section, 'section')
    else:
        section_tag = parent

    if section.type is not None:
        section_tag['data-type'] = section.type

    if section.type in TYPE_H_TAGS:
        h_tag_name = TYPE_H_TAGS[section.type]
    else:
        level = section.level

        if level < 1:
            level = 1
        elif level > 6:
            level = 6

        h_tag_name = f'h{level}'

    h_tag = section_tag.append_tag(h_tag_name)

    if section.number is not None:
        h_tag['data-number'] = section.number

    generate_component(h_tag, section.heading, collapse_paragraph=True)

    generate_component(section_tag, section.content)


def generate_toc(parent: Tag, toc: TableOfContents):
    nav_tag = append_component_tag(parent, toc, 'nav')
    nav_tag['data-type'] = 'toc'

    if toc.title is not None:
        nav_tag.append_tag('h1', text=toc.title)

    generate_toc_elements(nav_tag, toc.elements)


def generate_toc_elements(parent: Tag, elements: List[ElementReference]):
    if len(elements) == 0:
        return

    list_tag = parent.append_tag('ol')

    for element in elements:
        item_tag = list_tag.append_tag('li')

        a_tag = item_tag.append_tag('a', {
            'href': f'#{element.reference}',
            'data-type': 'xref',
        }, text=element.title)

        if element.number is not None:
            a_tag['data-number'] = element.number

        generate_toc_elements(item_tag, element.elements)


def generate_separator(parent: Tag, separator: Separator):
    sep = append_component_tag(parent, separator, 'hr')
    sep['data-level'] = separator.level


def generate_box(parent: Tag, box: ContentBox):
    box_tag = append_component_tag(parent, box, 'div')

    if box.style is not None:
        box_tag['data-type'] = box.style

    generate_component(box_tag, box.content, collapse_section=True)


def generate_image(parent: Tag, image: Image):
    img_tag = parent.append_tag('img')
    img_tag['src'] = image.src

    if image.alt:
        img_tag['alt'] = image.alt


def generate_function_call(parent: Tag, call: FunctionCall):
    if call.result is None:
        raise StxError(
            f'Function {call.key} was not processed.', call.location)

    generate_component(parent, call.result)


def generate_layout(parent: Tag, layout: Layout):
    container = parent.append_tag('div', {
        'data-type': 'layout',
        'data-dir': layout.direction,
    })

    for component in layout.components:
        item = container.append_tag('div', {
            'data-type': 'layout-item',
        })

        generate_component(item, component)


def append_component_tag(
        parent: Tag, component: Component, tag_name: str) -> Tag:
    for other_id in reversed(component.get_other_refs()):
        parent.append_tag('a', {'id': other_id}, text='')

    tag = parent.append_tag(tag_name)

    tag_id = component.get_main_ref()

    if tag_id is not None:
        tag['id'] = tag_id

    return tag



import string
from io import StringIO
from typing import Optional, List, Union

from stx.compiling.reading.content import Content
from stx.data_notation.values import Value, Token, Entry, Group, Empty
from stx.utils.stx_error import StxError

TOKEN_DELIMITER_CHAR = '`'
TOKEN_CHARS = string.ascii_letters + string.digits + '_-./'
TOKEN_ESCAPE_CHAR = '\\'
ENTRY_SEPARATOR_CHAR = ':'
GROUP_SEPARATOR_CHAR = ','
GROUP_BEGIN_CHAR = '('
GROUP_END_CHAR = ')'
EMPTY_CHAR = '~'
WHITESPACE_CHARS = ' \t\r\n'


def parse_value(content: Content) -> Value:
    values = parse_values(content)
    length = len(values)

    if length == 0:
        return Empty()
    elif length == 1:
        return values[0]
    return Group(values)


def parse_entry(content: Content) -> Entry:
    location = content.get_location()

    entry = try_parse_entry(content)

    if entry is None:
        raise StxError('Expected token or entry', location)

    return entry


def try_parse_entry(content: Content) -> Optional[Entry]:
    token_or_entry = try_parse_token_or_entry(content)

    if token_or_entry is None:
        return None

    return token_or_entry.to_entry()


def parse_values(content: Content) -> List[Value]:
    items = []

    can_be_none = True

    while True:
        loc0 = content.get_location()

        value = try_parse_item(content)

        if value is None:
            if can_be_none:
                content.go_back(loc0)
                break
            else:
                raise Exception('Expected to read a group item')

        items.append(value)

        loc0 = content.get_location()

        skip_void(content)

        c = content.peek()

        if c == GROUP_SEPARATOR_CHAR:
            content.move_next()

            skip_void(content)

            can_be_none = False
        else:
            content.go_back(loc0)
            break

    return items


def try_parse_item(
        content: Content,
        allow_entry_separator=True) -> Optional[Value]:
    c = content.peek()

    if c == EMPTY_CHAR:
        content.move_next()

        return Empty()

    group = try_parse_group(content)

    if group is not None:
        return group

    return try_parse_token_or_entry(content, allow_entry_separator)


def try_parse_group(content: Content):
    c = content.peek()

    if c == GROUP_BEGIN_CHAR:
        content.move_next()

        skip_void(content)

        items = parse_values(content)

        skip_void(content)

        c = content.peek()

        if c != GROUP_END_CHAR:
            raise StxError('Expected group end char', content.get_location())

        content.move_next()

        return Group(items)

    return None


def try_parse_token_or_entry(
        content: Content,
        allow_entry_separator=True) -> Union[Token, Entry, None]:
    text = try_parse_text(content)

    if text is None:
        return None

    loc0 = content.get_location()

    skip_void(content)

    c = content.peek()

    if c == ENTRY_SEPARATOR_CHAR and allow_entry_separator:
        content.move_next()

        skip_void(content)

        entry_name = text
        entry_value = try_parse_item(content, allow_entry_separator=False)

        if entry_value is None:
            raise StxError('Expected an entry value', content.get_location())

        return Entry(entry_name, entry_value)

    group = try_parse_group(content)

    if group is not None:
        return Entry(text, group)

    # go before skipping void
    content.go_back(loc0)

    return Token(text)


def try_parse_text(content: Content) -> Optional[str]:
    c = content.peek()

    if c in TOKEN_CHARS:
        out = StringIO()

        while True:
            out.write(c)

            content.move_next()

            c = content.peek()

            if c is None or c not in TOKEN_CHARS:
                break

        return out.getvalue()
    elif c == TOKEN_DELIMITER_CHAR:
        content.move_next()

        out = StringIO()

        while True:
            c = content.peek()

            if c is None:
                raise Exception(f'Expected {TOKEN_DELIMITER_CHAR}')
            elif c == TOKEN_DELIMITER_CHAR:
                content.move_next()
                break
            elif c == TOKEN_ESCAPE_CHAR:
                content.move_next()

                c = content.peek()

                if c is None:
                    raise Exception('Expected escaped char.')
                elif c in [TOKEN_ESCAPE_CHAR, TOKEN_DELIMITER_CHAR]:
                    out.write(c)

                    content.move_next()
                else:
                    # TODO add unicode support

                    raise Exception(f'Invalid escaped char: {c}')
            else:
                content.move_next()

                out.write(c)

        return out.getvalue()

    return None


def skip_void(content: Content):
    while True:
        c = content.peek()

        if c is None:
            break
        elif c in WHITESPACE_CHARS:
            content.move_next()
        else:
            break



from __future__ import annotations

from typing import Any, Dict, TextIO, List, Optional

from stx.outputs.html.escaping import write_text


class Element:

    def render(self, out: TextIO):
        raise NotImplementedError()


class Text(Element):

    def __init__(self, content: str):
        self.content = content

    def render(self, out: TextIO):
        if self.content is not None:
            write_text(self.content, out)


class Literal(Element):

    def __init__(self, content: str):
        self.content = content

    def render(self, out: TextIO):
        out.write(self.content)


class Tag(Element):

    def __init__(
            self,
            name: str,
            children: List[Element] = None,
            attributes: Dict[str, Any] = None):
        self.name = name
        self.preserve_spaces = False
        self._children: Optional[List[Element]] = children
        self._attributes: Optional[Dict[str, Any]] = attributes

    def __getitem__(self, key: str) -> Any:
        return self.attributes[key]

    def __setitem__(self, key: str, value: Any):
        self.attributes[key] = value

    def append_tag(
            self,
            name: str,
            attributes: Dict[str, Any] = None,
            *args,
            text: str = None,
            text_literal=False) -> Tag:

        if text is not None:
            if text_literal:
                children = [Literal(text)]
            else:
                children = [Text(text)]
        else:
            children = None

        tag = Tag(name, attributes=attributes, children=children)
        self.children.append(tag)
        return tag

    def append_text(self, value: str) -> Text:
        element = Text(value)
        self.children.append(element)
        return element

    def append_literal(self, value: str) -> Literal:
        element = Literal(value)
        self.children.append(element)
        return element

    @property
    def attributes(self) -> Dict[str, Any]:
        if self._attributes is None:
            self._attributes = {}
        return self._attributes

    @property
    def children(self) -> List[Element]:
        if self._children is None:
            self._children = []
        return self._children

    def render(self, out: TextIO):
        out.write('<')
        out.write(self.name)

        if self._attributes is not None and len(self._attributes) > 0:
            out.write(' ')

            for index, (key, value) in enumerate(self._attributes.items()):
                if index > 0:
                    out.write(' ')
                out.write(key)

                if value is not None:
                    out.write('=')
                    out.write('"')
                    write_text(value, out)
                    out.write('"')

        out.write('>')

        if self._children is not None and len(self._children) > 0:
            for child in self._children:
                child.render(out)

            out.write('</')
            out.write(self.name)
            out.write('>')



from __future__ import annotations

from typing import List, TextIO

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class PlainText(Component):

    def __init__(self, location: Location, text: str):
        self.location = location
        self.content = text

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.INLINE

    def is_rich(self) -> bool:
        return False

    def write_text(self, output: TextIO):
        output.write(self.content)

    def get_children(self) -> List[Component]:
        return []

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from io import StringIO
from typing import List, Optional

from stx.compiling.composer import Composer
from stx.compiling.marks import all_marks, escape_char, literal_area_mark
from stx.compiling.marks import attribute_special_mark, directive_special_mark
from stx.compiling.marks import code_begin_mark, code_end_mark
from stx.compiling.marks import container_area_begin_mark
from stx.compiling.marks import container_area_end_mark
from stx.compiling.marks import container_begin_mark, container_end_mark
from stx.compiling.marks import d_quote_begin_mark, d_quote_end_mark
from stx.compiling.marks import deleted_begin_mark, deleted_end_mark
from stx.compiling.marks import ellipsis_single_mark
from stx.compiling.marks import emphasized_begin_mark, emphasized_end_mark
from stx.compiling.marks import function_begin_mark, function_end_mark
from stx.compiling.marks import heading_block_marks, header_row_block_mark, \
    not_inline_marks, inline_marks, link_ref_begin_mark, link_ref_end_mark
from stx.compiling.marks import link_text_begin_mark, link_text_end_mark
from stx.compiling.marks import normal_row_block_mark, cell_block_mark
from stx.compiling.marks import ordered_item_block_mark, section_levels
from stx.compiling.marks import post_caption_block_mark
from stx.compiling.marks import pre_caption_block_mark
from stx.compiling.marks import s_quote_begin_mark, s_quote_end_mark
from stx.compiling.marks import strong_begin_mark, strong_end_mark
from stx.compiling.marks import unordered_item_block_mark
from stx.compiling.reading.content import Content
from stx.compiling.reading.location import Location
from stx.compiling.reading.reader import Reader
from stx.components import Component, Section, Composite, Table, TableRow, \
    Paragraph, DisplayMode
from stx.components import FunctionCall
from stx.components import LinkText, StyledText, PlainText, Literal, ListBlock
from stx.data_notation.parsing import parse_entry, skip_void, try_parse_entry
from stx.data_notation.values import Value
from stx.document import Document
from stx.outputs import make_output_action
from stx.utils.closeable import Closeable
from stx.utils.debug import see
from stx.utils.files import resolve_include_files
from stx.utils.stx_error import StxError

PASS = 0
EXIT = 1
CONSUMED = 2


class CTX:

    def __init__(self, document: Document, reader: Reader):
        self.reader = reader
        self.document = document
        self.composer = Composer()
        self.stop_mark_stack = []
        self.section_stack: List[Section] = []

    def get_parent_section(self) -> Optional[Section]:
        if len(self.section_stack) > 0:
            return self.section_stack[-1]
        return None

    @property
    def stop_mark(self) -> Optional[str]:  # TODO rename to stop token
        if len(self.stop_mark_stack) == 0:
            return None
        return self.stop_mark_stack[-1]

    def using_stop_mark(self, char: str) -> Closeable:
        def enter_action():
            self.stop_mark_stack.append(char)

        def exit_action():
            self.stop_mark_stack.pop()

        return Closeable(enter_action, exit_action)


def capture(ctx: CTX):
    ctx.document.content = capture_component(
        ctx,
        indentation=0,
        breakable=False)

    if ctx.reader.active():
        raise StxError('Unexpected content.', ctx.reader.get_location())

    # TODO improve error messages
    if len(ctx.composer.stack) > 0:
        raise StxError('unclosed components')
    elif not ctx.composer.attributes_buffer.empty():
        raise StxError(
            f'not consumed attributes: {ctx.composer.attributes_buffer}')
    elif len(ctx.composer.pre_captions) > 0:
        raise StxError(
            'not consumed pre captions: ',
            ctx.composer.pre_captions[0].location)


def capture_component(
        ctx: CTX,
        indentation: int,
        breakable=True) -> Component:
    location = ctx.reader.get_location()

    ctx.composer.push()

    while ctx.reader.active():
        content = ctx.reader.get_content()

        while content.consume_empty_line():
            pass

        if not content.consume_indentation(indentation):
            break
        elif content.test(ctx.stop_mark):
            break
        elif content.peek() is None:
            break

        location = content.get_location()

        mark = content.read_mark(not_inline_marks)

        signal = (
                parse_section(ctx, mark, location, content, indentation)
                or
                parse_list(ctx, mark, location, content)
                or
                parse_table(ctx, mark, location, content)
                or
                parse_caption(ctx, mark, content)
                or
                parse_literal(ctx, mark, location, content, indentation)
                or
                parse_container(ctx, mark, location, content, indentation)
                or
                parse_attribute(ctx, mark, content)
                or
                parse_directive(ctx, mark, location)
        )

        if signal == PASS:
            parse_inline_component(ctx, mark, location, content, indentation)
        elif signal == CONSUMED:
            pass
        elif signal == EXIT:
            break
        else:
            raise AssertionError(f'Illegal signal: {signal}')

    components = ctx.composer.pop()

    if len(components) == 0:
        # TODO Blank component
        return Composite(location, [])
    elif len(components) == 1:
        return components[0]

    return Composite(location, components)


def parse_section(
        ctx: CTX,
        mark: str,
        location: Location,
        content: Content,
        before_mark_indentation: int) -> int:
    if mark not in heading_block_marks:
        return PASS

    parent_section = ctx.get_parent_section()
    section_level = section_levels[mark]

    if (parent_section is not None
            and parent_section.level >= section_level):
        content.go_back(location)
        return EXIT

    # TODO is this ok?
    skip_void(content)

    after_mark_indentation = content.column

    section = Section(location, section_level)

    ctx.composer.add(section)

    ctx.section_stack.append(section)

    section.heading = capture_component(ctx, after_mark_indentation, True)
    section.content = capture_component(ctx, before_mark_indentation, True)

    ctx.section_stack.pop()

    return CONSUMED


def parse_list(
        ctx: CTX,
        mark: str,
        location: Location,
        content: Content) -> int:
    if mark == ordered_item_block_mark:
        ordered = True
    elif mark == unordered_item_block_mark:
        ordered = False
    else:
        return PASS

    list_block = ctx.composer.get_last_component()

    if not isinstance(list_block, ListBlock):
        list_block = ListBlock(location, ordered)

        ctx.composer.add(list_block)

    # TODO is this ok?
    skip_void(content)

    indentation = content.column

    list_item = capture_component(ctx, indentation, True)

    list_block.items.append(list_item)

    return CONSUMED


def parse_table(
        ctx: CTX,
        mark: str,
        location: Location,
        content: Content) -> int:
    if mark == header_row_block_mark:
        header = True
        reuse_row = False
    elif mark == normal_row_block_mark:
        header = False
        reuse_row = False
    elif mark == cell_block_mark:
        header = False
        reuse_row = True
    else:
        return PASS

    table = ctx.composer.get_last_component()

    if not isinstance(table, Table):
        table = Table(location)

        ctx.composer.add(table)

    row = table.get_last_row() if reuse_row else None

    if row is None:
        row = TableRow(location, header)

        table.rows.append(row)

    # TODO is this ok?
    skip_void(content)

    indentation0 = content.column
    indentation = indentation0

    while True:
        with ctx.using_stop_mark(cell_block_mark):
            cell = capture_component(ctx, indentation, True)

            row.cells.append(cell)

        if not ctx.reader.active():
            break

        content = ctx.reader.get_content()

        loc0 = content.get_location()

        # Consume indentation when it is the beginning of the line
        if content.column == 0:
            if content.read_spaces(indentation0) < indentation0:
                content.go_back(loc0)
                break

        if content.peek() == cell_block_mark:
            content.move_next()
            content.read_spaces()

            indentation = content.column
        else:
            break

    return CONSUMED


def parse_caption(ctx: CTX, mark: str, content: Content) -> int:
    if mark == pre_caption_block_mark:
        pre_mode = True
    elif mark == post_caption_block_mark:
        pre_mode = False
    else:
        return PASS

    # TODO is this ok?
    skip_void(content)

    caption = capture_component(ctx, content.column, True)

    if pre_mode:
        ctx.composer.push_pre_caption(caption)
    else:
        ctx.composer.push_post_caption(caption)

    return CONSUMED


def parse_literal(
        ctx: CTX, mark: str, location: Location, content: Content,
        indentation_before_mark: int) -> int:
    if mark != literal_area_mark:
        return PASS

    content.read_spaces()

    function_location = content.get_location()

    function = try_parse_entry(content)

    content.expect_end_of_line()

    out = StringIO()

    while True:
        line = content.read_line(indentation_before_mark)

        if line is None:
            raise StxError(f'Expected: {mark}', content.get_location())
        elif line.startswith(escape_char):
            line = line[1:]  # remove escape char

            if not line.startswith(mark) and not line.startswith(escape_char):
                raise StxError(f'Invalid escaped sequence, expected:'
                               f' {see(mark)} or {see(escape_char)}.')
        elif line.rstrip() == mark:
            break

        out.write(line)

    text = out.getvalue()

    if function is not None:
        component = FunctionCall(
            function_location,
            inline=False,
            key=function.name,
            options=function.value,
            argument=Literal(location, text),
        )
    else:
        component = Literal(location, text)

    ctx.composer.add(component)

    return CONSUMED


def parse_container(
        ctx: CTX, mark: str, location: Location, content: Content,
        indentation_before_mark: int) -> int:
    if mark != container_area_begin_mark:
        return PASS

    content.read_spaces()

    function = try_parse_entry(content)

    content.expect_end_of_line()

    with ctx.using_stop_mark(container_area_end_mark):
        component = capture_component(ctx, indentation_before_mark)

    content.pull(container_area_end_mark)
    content.expect_end_of_line()

    if function is not None:
        component = FunctionCall(
            location,
            inline=False,
            key=function.name,
            options=function.value,
            argument=component,
        )

    ctx.composer.add(component)

    return CONSUMED


def parse_attribute(ctx: CTX, mark: str, content: Content) -> int:
    if mark != attribute_special_mark:
        return PASS

    entry = parse_entry(content)

    content.expect_end_of_line()

    ctx.composer.push_attribute(entry.name, entry.value)

    return CONSUMED


def parse_directive(ctx: CTX, mark: str, location: Location) -> int:
    if mark != directive_special_mark:
        return PASS

    content = ctx.reader.get_content()
    file_path = content.file_path

    entry = parse_entry(content)

    key = entry.name
    value = entry.value

    if content.column > 0:
        content.expect_end_of_line()

    if key == 'title':
        ctx.document.title = value.to_str()
    elif key == 'author':
        ctx.document.author = value.to_str()
    elif key == 'format':
        ctx.document.format = value.to_str()
    elif key == 'encoding':
        ctx.document.encoding = value.to_str()
    elif key == 'stylesheets':
        ctx.document.stylesheets = value.to_list()
    elif key == 'include':
        process_import(ctx, location, file_path, value)
    elif key == 'output':
        process_output(ctx, location, value)
    else:
        raise StxError(f'Unsupported directive: {key}')

    return CONSUMED


def process_import(
        ctx: CTX,
        location: Location,
        file_path: str,
        include_path: Value):
    file_paths = resolve_include_files(include_path.to_str(), file_path)

    ctx.reader.push_files(file_paths)


def process_output(ctx: CTX, location: Location, value: Value):
    action = make_output_action(ctx.document, location, value)

    ctx.document.actions.append(action)


def parse_inline_component(
        ctx: CTX,
        mark: str,
        location: Location,
        content: Content,
        indentation: int):
    inlines = parse_inline(ctx, location, content, indentation)

    if len(inlines) == 0:
        # TODO return empty component
        raise StxError('Missing content.', location)
    elif len(inlines) == 1 and inlines[0].display_mode == DisplayMode.BLOCK:
        component = inlines[0]
    else:
        display_mode = DisplayMode.compute_display_mode(inlines)

        if display_mode == DisplayMode.BLOCK:
            component = Composite(location, inlines)
        else:
            component = Paragraph(location, inlines)

    ctx.composer.add(component)


def parse_inline(
        ctx: CTX, location: Location,
        content: Content, indentation: int) -> List[Component]:
    ctx.composer.push()

    try:
        while not content.halted():
            if content.test(ctx.stop_mark):
                break

            location = content.get_location()

            mark = content.read_mark(inline_marks)

            signal = (
                parse_inline_function(ctx, mark, location, content)
                or
                parse_inline_container(
                    ctx, mark, location, content, indentation)
                or
                parse_inline_style(ctx, mark, location, content, indentation)
                or
                parse_inline_link(ctx, mark, location, content, indentation)
                or
                parse_inline_token(ctx, mark, location, content, indentation)
                or
                parse_inline_text(ctx, mark, location, content, indentation)
            )

            if signal == PASS:
                raise StxError(f'Not implemented mark: {mark}')
            elif signal == CONSUMED:
                pass
            elif signal == EXIT:
                break
            else:
                raise AssertionError(f'Illegal signal: {signal}')
    except StxError as e:
        raise StxError('Error parsing inline content.', location) from e

    return ctx.composer.pop()


def parse_inline_function(
        ctx: CTX, mark: str, location: Location, content: Content) -> int:
    if mark != function_begin_mark:
        return PASS

    skip_void(content)

    function_location = content.get_location()

    function = parse_entry(content)

    skip_void(content)

    if not content.pull(function_end_mark):
        raise StxError(f'Expected mark: {function_end_mark}')

    ctx.composer.add(
        FunctionCall(
            function_location,
            inline=True,
            key=function.name,
            options=function.value))

    return CONSUMED


def parse_inline_container(
        ctx: CTX, mark: str, location: Location,
        content: Content, indentation: int) -> int:
    if mark != container_begin_mark:
        return PASS

    with ctx.using_stop_mark(container_end_mark):
        # It uses the original indentation
        #   so the paragraph can be continued.
        contents = parse_inline(
            ctx, content.get_location(), content, indentation)

    if not content.pull(container_end_mark):
        raise StxError(f'Expected mark: {container_end_mark}')

    if not content.pull(function_begin_mark):
        raise StxError(f'Expected mark: {function_begin_mark}')

    skip_void(content)

    function_location = content.get_location()

    function = parse_entry(content)

    skip_void(content)

    if not content.pull(function_end_mark):
        raise StxError(f'Expected mark: {function_end_mark}')

    ctx.composer.add(
        FunctionCall(
            function_location,
            inline=True,
            key=function.name,
            options=function.value,
            argument=Composite(location, contents)))

    return CONSUMED


def parse_inline_style(
        ctx: CTX, mark: str, location: Location,
        content: Content, indentation: int) -> int:
    if mark == strong_begin_mark:
        end_mark = strong_end_mark
        style = 'strong'
    elif mark == emphasized_begin_mark:
        end_mark = emphasized_end_mark
        style = 'emphasized'
    elif mark == code_begin_mark:
        end_mark = code_end_mark
        style = 'code'
    elif mark == deleted_begin_mark:
        end_mark = deleted_end_mark
        style = 'deleted'
    elif mark == d_quote_begin_mark:
        end_mark = d_quote_end_mark
        style = 'double-quote'
    elif mark == s_quote_begin_mark:
        end_mark = s_quote_end_mark
        style = 'single-quote'
    else:
        return PASS

    with ctx.using_stop_mark(end_mark):
        # It uses the original indentation
        #   so the paragraph can be continued.
        contents = parse_inline(
            ctx, content.get_location(), content, indentation)

    if not content.pull(end_mark):
        raise StxError(f'Expected mark: {end_mark}')

    ctx.composer.add(
        StyledText(location, contents, style)
    )

    return CONSUMED


def parse_inline_link(
        ctx: CTX, mark: str, location: Location,
        content: Content, indentation: int) -> int:
    if mark != link_text_begin_mark:
        return PASS

    with ctx.using_stop_mark(link_text_end_mark):
        # It uses the original indentation
        #   so the paragraph can be continued.
        contents = parse_inline(
            ctx, content.get_location(), content, indentation)

    if not content.pull(link_text_end_mark):
        raise StxError(f'Expected mark: {link_text_end_mark}')

    if content.pull(link_ref_begin_mark):
        out = StringIO()

        while not content.pull(link_ref_end_mark):
            c = content.peek()

            if c is None:
                raise StxError(
                    f'Expected {link_ref_end_mark}', content.get_location())

            out.write(c)
            content.move_next()

        reference = out.getvalue()
    else:
        reference = None

    ctx.composer.add(LinkText(location, contents, reference))

    return CONSUMED


def parse_inline_token(
        ctx: CTX, mark: str, location: Location,
        content: Content, indentation: int) -> int:
    if mark == ellipsis_single_mark:
        text = '\u2026'
    else:
        return PASS

    ctx.composer.add(
        PlainText(location, text)
    )

    return CONSUMED


def parse_inline_text(
        ctx: CTX, mark: str, location: Location,
        content: Content, indentation: int) -> int:
    if mark is not None:
        return PASS

    out = StringIO()

    completed = False

    while content.peek() is not None:
        # Check if the text is broken by an inline or stop mark
        if content.test_any(inline_marks):
            break
        elif content.test(ctx.stop_mark):
            break

        c = content.peek()

        if c == '\n':
            out.write(c)
            content.move_next()

            # Check if the text is completed by an empty line
            if content.consume_empty_line():
                completed = True
                break

            loc0 = content.get_location()

            spaces = content.read_spaces(indentation)

            # Check if the text is completed by indentation change
            if spaces < indentation:
                content.go_back(loc0)
                completed = True
                break

            # Check if the text is completed by a non-inline mark
            if content.test_any(not_inline_marks):
                content.go_back(loc0)
                completed = True
                break
        elif c == escape_char:
            content.move_next()

            escaped_mark = content.pull_any(all_marks)
            if escaped_mark is not None:
                out.write(escaped_mark)
            elif content.pull(ctx.stop_mark):
                out.write(ctx.stop_mark)
            elif content.pull(escape_char):
                out.write(escape_char)
            else:
                raise StxError('invalid escaped char')
        else:
            out.write(c)
            content.move_next()

    text = out.getvalue()

    if text == '':
        return EXIT

    ctx.composer.add(
        PlainText(location, text)
    )

    if completed:
        return EXIT
    return CONSUMED



import string
from typing import Set, Optional

from stx import logger
from stx.components import Component, Section, Table, Figure, LinkText
from stx.document import Document
from stx.utils.debug import see

REF_LENGTH_HINT = 40


# TODO refactor `ref` attribute, auto-generation,
#  normalization and validation.


def auto_generate_special_references(document: Document):
    roots = list()
    refs = set()

    if document.content is not None:
        roots.append(document.content)

    for root in roots:
        normalize_references(root)

    for root in roots:
        collect_references(root, refs)

    for root in roots:
        register_missing_references(root, refs)


def validate_references(document: Document):
    roots = list()
    refs = set()

    if document.content is not None:
        roots.append(document.content)

    for root in roots:
        collect_references(root, refs)

    for root in roots:
        normalize_and_report_invalid_links(root, refs)


def collect_references(root: Component, refs: Set[str]):
    for component in root.walk():
        for ref in component.get_refs():
            if ref in refs:
                raise component.error(f'Reference already taken: {ref}')
            else:
                refs.add(ref)


def normalize_references(root: Component):
    for component in root.walk():
        wild_refs = component.get_refs()

        if len(wild_refs) > 0:
            component.ref = [
                make_ref(wild_ref) for wild_ref in wild_refs
            ]


def register_missing_references(root: Component, refs: Set[str]):
    for component in root.walk():
        if isinstance(component, Section):
            register_section_reference(component, refs)
        elif isinstance(component, Table):
            register_table_reference(component, refs)
        elif isinstance(component, Figure):
            register_figure_reference(component, refs)


def normalize_and_report_invalid_links(root: Component, refs: Set[str]):
    for component in root.walk():
        if isinstance(component, LinkText):
            if component.reference is None:
                # Generate reference from the text
                component.reference = make_ref(component.get_text())
            elif component.is_internal():
                # Normalize wild reference
                component.reference = make_ref(component.reference)
            else:
                # Do not validate this link
                continue

            if component.reference not in refs:
                component.invalid = True
                logger.warning(
                    f'Invalid link: {see(component.reference, None)}',
                    component.location)


def register_section_reference(section: Section, refs: Set[str]):
    ref = generate_component_ref(section.heading, refs)

    section.add_ref(ref)
    refs.add(ref)


def register_table_reference(table: Table, refs: Set[str]):
    if table.caption is not None:
        ref = generate_component_ref(table.caption, refs)
    else:
        ref = generate_component_ref(table, refs)

    table.add_ref(ref)
    refs.add(ref)


def register_figure_reference(figure: Figure, refs: Set[str]):
    ref = generate_component_ref(figure.caption, refs)

    figure.add_ref(ref)
    refs.add(ref)


def generate_component_ref(component: Component, refs: Set[str]) -> str:
    count = 0

    plain_text = component.get_text()

    while True:
        ref = make_ref(plain_text, REF_LENGTH_HINT, count)

        if ref not in refs:
            break

        count += 1

    return ref


def make_ref(
        base: str,
        length_hint: Optional[int] = None,
        count: Optional[int] = None) -> str:
    result = []

    # Important generate it in lowercase and trimmed
    for c in base.strip().lower():
        if c in string.ascii_lowercase or c in string.digits:
            result.append(c)
        elif length_hint is not None and len(result) >= length_hint:
            break
        elif len(result) == 0 or result[-1] != '-':
            result.append('-')

    ref = ''.join(result)

    if count is not None and count > 0:
        if ref.endswith('-'):
            ref += str(count)
        else:
            ref += f'-{count}'

    return ref



from stx.functions import registry
from stx.components import Component, FunctionCall
from stx.document import Document
from stx.utils.stx_error import StxError
from stx.utils.debug import see


def resolve_document(document: Document) -> int:
    count = 0

    if document.content is not None:
        count += resolve_component(document, document.content)

    return count


def resolve_component(document: Document, component: Component) -> int:
    count = 0
    children = component.get_children()

    # Resolve children first
    for child in children:
        count += resolve_component(document, child)

    # Resolve function then
    if isinstance(component, FunctionCall):
        count += resolve_function_call(document, component)

    return count


def resolve_function_call(document: Document, call: FunctionCall):
    # Check if it is already resolved
    if call.result is not None:
        return 0

    processor = registry.get(call.key)

    if processor is None:
        raise StxError(
            f'Function not found: {see(call.key)}', call.location)

    call.result = processor(document, call)

    if call.result is None:
        raise call.error('Function call did not produce a component.')

    return 1 + resolve_component(document, call.result)



from typing import Callable, Any


class Closeable:

    def __init__(
            self,
            enter_action: Callable[[], Any],
            exit_action: Callable[[], Any]):
        self.enter_action = enter_action
        self.exit_action = exit_action

    def __enter__(self):
        self.enter_action()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            raise exc_val

        self.exit_action()



from typing import List


class HtmlTheme:

    def __init__(self):
        self.head_styles: List[str] = []
        self.head_scripts: List[str] = []
        self.body_scripts: List[str] = []


class NullHtmlTheme(HtmlTheme):

    pass



from io import StringIO
from typing import Any, Optional

DEFAULT_MAX_LENGTH = 40


def quote_string(value: str, max_length: Optional[int]) -> str:
    out = StringIO()

    out.write('`')

    actual_length = 0

    for c in value:
        if max_length is not None and actual_length >= max_length:
            break
        elif c == '\\':
            out.write('\\\\')
        elif c == '\n':
            out.write('\\n')
        elif c == '\r':
            out.write('\\r')
        elif c == '\t':
            out.write('\\t')
        else:
            unicode = ord(c)

            if 0x20 <= unicode <= 0x7E:
                out.write(c)
            else:
                out.write(f'\\({unicode})')

        actual_length += 1

    if actual_length < len(value):
        out.write('\u2026')

    out.write('`')

    return out.getvalue()


def see(value: Any, limit=DEFAULT_MAX_LENGTH) -> str:
    if isinstance(value, str):
        return quote_string(str(value), limit)
    elif isinstance(value, int):
        return str(value)
    elif isinstance(value, dict):
        return str(value)

    return f'{type(value).__name__}<{hash(value)}>'



from __future__ import annotations

from typing import List
from typing import Optional

from stx.action import Action
from stx.components import Component


class Document:

    def __init__(self, source_file):
        self.source_file = source_file
        self.title: Optional[str] = None
        self.author: Optional[str] = None
        self.format: Optional[str] = None
        self.encoding: Optional[str] = None
        self.header: Optional[Component] = None
        self.content: Optional[Component] = None
        self.footer: Optional[Component] = None
        self.stylesheets: List[str] = []
        self.actions: List[Action] = []



from __future__ import annotations

from typing import List, TextIO

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.stx_error import StxError
from ..utils.tracked_dict import TrackedDict


DIR_ROW = 'row'
DIR_COLUMN = 'col'


class Layout(Component):

    def __init__(
            self,
            location: Location,
            components: List[Component],
            direction: str):
        self.location = location
        self.components = components

        if direction not in [DIR_ROW, DIR_COLUMN]:
            raise StxError(f'Not allowed direction: {direction}', location)

        self.direction = direction

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.compute_display_mode(self.components)

    def is_rich(self) -> bool:
        for c in self.components:
            if c.is_rich():
                return True
        return False

    def write_text(self, output: TextIO):
        for component in self.components:
            component.write_text(output)

    def get_children(self) -> List[Component]:
        return self.components

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from stx import logger
from stx.functions import utils
from stx.components import Component, FunctionCall, Literal
from stx.document import Document
from stx.utils.files import resolve_sibling
from stx.utils.stx_error import StxError


def resolve_embed(document: Document, call: FunctionCall) -> Component:
    options = utils.make_options_dict(call, 'src')

    src = options.pop('src', None)

    utils.check_unknown_options(options, call)

    if src is None:
        raise StxError(f'Missing `src` parameter in embed.', call.location)

    src = resolve_sibling(document.source_file, src)

    logger.info(f'Embedding: {src}')

    with open(src, 'r', encoding='UTF-8') as f:
        text = f.read()

    # TODO add support for more type of files

    return Literal(call.location, text, source=src)



from stx import resources
from stx.outputs.html.themes import HtmlTheme


classic_theme = HtmlTheme()

classic_theme.head_styles.extend([
    resources.get_text('classic_html/layout.css'),
    resources.get_text('classic_html/style.css'),
])

classic_theme.body_scripts.extend([
    resources.get_text('classic_html/jquery-3.4.1.min.js'),
    resources.get_text('classic_html/create-layout.js'),
    resources.get_text('classic_html/toc-links.js'),
])



from stx.functions import utils
from stx.components import Component, FunctionCall, CustomText
from stx.document import Document


def resolve_custom_style(document: Document, call: FunctionCall) -> Component:
    if call.argument is None:
        raise call.error(
            'Custom style function requires a captured component.')

    options = utils.make_options_dict(call, key_for_str='style')
    contents = utils.make_component_list(call.argument)

    custom_style = options.pop('style')

    utils.check_unknown_options(options, call)

    return CustomText(call.location, contents, custom_style)



from stx import logger
from stx.functions import utils
from stx.components import Component, Image, FunctionCall
from stx.document import Document
from stx.utils.stx_error import StxError


def resolve_image(document: Document, call: FunctionCall) -> Component:
    options = utils.make_options_dict(call, 'src')

    src = options.pop('src', None)
    alt = options.pop('alt', None)

    utils.check_unknown_options(options, call)

    if src is None:
        raise StxError(f'Missing `src` parameter in image.', call.location)
    elif not alt:
        logger.warning('Image without `alt`', call.location)

    return Image(call.location, src, alt)



from typing import Dict, Type

from stx import logger
from stx.compiling.reading.location import Location
from stx.data_notation.values import Value, Empty
from stx.document import Document
from stx.outputs.html.output import HtmlOutputAction
from stx.outputs.json.output import JsonOutputAction
from stx.outputs.output_action import OutputAction, OutputStdOut
from stx.outputs.output_action import OutputTarget
from stx.utils.stx_error import StxError
from stx.utils.debug import see

_handler_types: Dict[str, Type[OutputAction]] = {}


def register_type(
        format_key: str, handler_type: Type[OutputAction], override=False):
    if not override and format_key in _handler_types:
        raise StxError(f'Output handler already registered: {format_key}')

    _handler_types[format_key] = handler_type


def get_type(format_key: str) -> Type[OutputAction]:
    if format_key not in _handler_types:
        raise StxError(f'Format not supported: {format_key}')
    return _handler_types[format_key]


def make_output_action(
        document: Document,
        location: Location,
        arguments: Value) -> OutputAction:
    format_value = arguments.try_token()

    if format_value is not None:
        actual_format = format_value.to_str()
        actual_target = OutputStdOut()
        actual_theme = None
        actual_options = Empty()
    else:
        args_map = arguments.to_map()

        format_value = args_map.pop('format', None)
        target_value = args_map.pop('target', None)
        theme_value = args_map.pop('theme', None)
        options_value = args_map.pop('options', None)

        if len(args_map) > 0:
            for unknown_key in args_map.keys():
                logger.warning(
                    f'Unknown output argument: {see(unknown_key)}',
                    location)

        if format_value is not None:
            actual_format = format_value.to_str()
        else:
            raise StxError('Expected output format.', location)

        if target_value is not None:
            actual_target = OutputTarget.make(
                document, target_value.to_str())
        else:
            actual_target = OutputStdOut()

        if theme_value is not None:
            actual_theme = theme_value.to_str()
        else:
            actual_theme = None

        if options_value is not None:
            actual_options = options_value
        else:
            actual_options = Empty()

    handler_type = get_type(actual_format)

    return handler_type(
        document=document,
        location=location,
        format_key=actual_format,
        target=actual_target,
        options=actual_options,
        theme=actual_theme,
    )


# Built-in renderers
register_type('html', HtmlOutputAction)
register_type('json', JsonOutputAction)



import pkg_resources


def get_text(name: str) -> str:
    try:
        content = pkg_resources.resource_string(__name__, 'data/' + name)
        text = content.decode('utf-8')
        return text
    except FileNotFoundError:
        raise Exception(f'Resource not found: {name}')



from __future__ import annotations

from typing import List, TextIO, Optional

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class Literal(Component):

    def __init__(
            self,
            location: Location,
            text: str,
            lang: Optional[str] = None,  # TODO remove this field
            source: Optional[str] = None):
        self.location = location
        self.text = text
        self.lang = lang
        self.source = source

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.DEFAULT

    def is_rich(self) -> bool:
        return False

    def write_text(self, output: TextIO):
        output.write(self.text)

    def get_children(self) -> List[Component]:
        return []

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from typing import Generic, TypeVar, Optional, Dict, Set

KT = TypeVar('KT')
VT = TypeVar('VT')


class TrackedDict(Generic[KT, VT]):

    def __init__(self, source: Dict[KT, VT]):
        self._source = source
        self._active_keys: Set[KT] = set()

    def get(self, key: KT) -> Optional[VT]:
        self._active_keys.add(key)
        return self._source.get(key, None)

    def put(self, key: KT, value: VT):
        self._source[key] = value

    def clear(self):
        self._source.clear()

    def empty(self) -> bool:
        return len(self._source) == 0

    def unknown_keys(self) -> Set[KT]:
        return self._source.keys() - self._active_keys

    def active_keys(self) -> Set[KT]:
        return set(self._active_keys)

    def missing_keys(self) -> Set[KT]:
        return self._active_keys - self._source.keys()

    def keys(self):
        return self._source.keys()



from typing import List, Optional

# Block marks

heading6_block_mark = '======'  # TODO rename to section
heading5_block_mark = '====='
heading4_block_mark = '===='
heading3_block_mark = '==='
heading2_block_mark = '=='
heading1_block_mark = '='
unordered_item_block_mark = '-'
ordered_item_block_mark = '.'
header_row_block_mark = '|='
normal_row_block_mark = '|-'
cell_block_mark = '|'
pre_caption_block_mark = '::'
post_caption_block_mark = ':^'

# Wrapping marks

function_begin_mark = '<'
function_end_mark = '>'
container_begin_mark = '{'
container_end_mark = '}'
strong_begin_mark = '*'
strong_end_mark = '*'
emphasized_begin_mark = '_'
emphasized_end_mark = '_'
code_begin_mark = '`'
code_end_mark = '`'
deleted_begin_mark = '~~'
deleted_end_mark = '~~'
d_quote_begin_mark = '""'
d_quote_end_mark = '""'
s_quote_begin_mark = '\'\''
s_quote_end_mark = '\'\''
link_text_begin_mark = '['
link_text_end_mark = ']'
link_ref_begin_mark = '('
link_ref_end_mark = ')'

# Single Marks

ellipsis_single_mark = '...'

# Area marks

literal_area_mark = '+++'
container_area_begin_mark = '{{{'
container_area_end_mark = '}}}'

# Special Marks

attribute_special_mark = '@'
directive_special_mark = '#'
exit_special_mark = '%'

escape_char = '\\'


# Summaries (the order matters for parsing)

heading_block_marks = [
    heading6_block_mark,
    heading5_block_mark,
    heading4_block_mark,
    heading3_block_mark,
    heading2_block_mark,
    heading1_block_mark,
]

section_levels = {
    heading1_block_mark: 1,
    heading2_block_mark: 2,
    heading3_block_mark: 3,
    heading4_block_mark: 4,
    heading5_block_mark: 5,
    heading6_block_mark: 6,
}

style_token_map = {
    '*': 'strong',
    '_': 'emphasized',
    '`': 'code',
}


def get_matching_mark(text: str, valid_marks: List[str]) -> Optional[str]:
    for mark in valid_marks:
        if text.startswith(mark):
            return mark

    return None


not_inline_marks = [
    heading6_block_mark,
    heading5_block_mark,
    heading4_block_mark,
    heading3_block_mark,
    heading2_block_mark,
    heading1_block_mark,

    unordered_item_block_mark,
    ordered_item_block_mark,
    header_row_block_mark,
    normal_row_block_mark,
    cell_block_mark,
    pre_caption_block_mark,
    post_caption_block_mark,

    literal_area_mark,
    container_area_begin_mark,
    container_area_end_mark,

    attribute_special_mark,
    directive_special_mark,
    exit_special_mark,
]


inline_marks = [
    function_begin_mark,
    container_begin_mark,
    strong_begin_mark,
    emphasized_begin_mark,
    code_begin_mark,
    deleted_begin_mark,
    d_quote_begin_mark,
    s_quote_begin_mark,
    link_text_begin_mark,
    ellipsis_single_mark,
]

all_marks = not_inline_marks + inline_marks

mark_max_length = max(len(mark) for mark in all_marks)



import setuptools


def read_content(file: str) -> str:
    with open(file, 'r') as f:
        return f.read().strip()


version_number = read_content('version.txt')

# development_status = 'Development Status :: 3 - Alpha'
development_status = 'Development Status :: 4 - Beta'
# development_status = 'Development Status :: 5 - Production/Stable'


setuptools.setup(
    name='stx',
    version=version_number,
    author='Sergio Pedraza',
    author_email='sergio.uriel.ph@gmail.com',
    description='Python implementation of the STX Language',
    long_description=read_content('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/stx-lang/python-stx',
    packages=setuptools.find_packages(),
    package_data={'stx': ['resources/data/*/*']},
    python_requires='>=3.8',
    license='MIT',
    install_requires=read_content('requirements-main.txt'),
    entry_points={
        'console_scripts': [
            'stx=stx.app:cli'
        ]
    },
    # See: https://pypi.org/classifiers/
    classifiers=[
        development_status,
        'Topic :: Documentation',
        'Topic :: Text Processing :: Markup',
        'Topic :: Software Development :: Documentation',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import json
from typing import TextIO

from stx.outputs.json.serializer import document_to_json
from stx.outputs.output_action import OutputAction


class JsonOutputAction(OutputAction):

    def dump(self, out: TextIO):
        d = document_to_json(self.document)

        # TODO implement options
        indent = 2

        json.dump(d, out, indent=indent)



from typing import Any

from stx.compiling.reading.content import Content
from stx.data_notation.parsing import parse_value


def parse(text: str) -> Any:
    source = Content(text, 'test')

    return parse_value(source).to_any()


def test_parse_value():
    assert parse('~') is None
    assert parse('abc') == 'abc'
    assert parse('`abc`') == 'abc'
    assert parse('a:1') == {'a': '1'}
    assert parse('nothing:~') == {'nothing': None}
    assert parse('~, ~, ~') == [None, None, None]
    assert parse('a:1, b:2') == {'a': '1', 'b': '2'}
    assert parse('a:1, b:(2, 3, 4), c:5') == {
        'a': '1', 'b': ['2', '3', '4'], 'c': '5'}



from typing import Generic, TypeVar, List, Optional

T = TypeVar('T')


class Stack(Generic[T]):

    def __init__(self):
        self.items: List[T] = []

    def empty(self) -> bool:
        return len(self.items) == 0

    def push(self, item: T):
        self.items.append(item)

    def pop(self) -> T:
        return self.items.pop()

    def peek(self, default: Optional[T] = None) -> Optional[T]:
        if len(self.items) > 0:
            return self.items[-1]

        return default



from stx.functions import utils
from stx.components import FunctionCall, Component, Layout
from stx.document import Document


def layout_function(document: Document, call: FunctionCall) -> Component:
    if call.argument is None:
        raise call.error('Layouts require a captured component.')

    components = utils.make_component_list(call.argument)

    options = utils.make_options_dict(call, key_for_str='dir')

    direction = options.pop('dir', None)

    utils.check_unknown_options(options, call)

    return Layout(call.location, components, direction)



from typing import Optional, Union, List

from stx import logger
from stx.components import (
    FunctionCall, Component, Composite, Paragraph)
from stx.data_notation.values import Value
from stx.utils.stx_error import StxError
from stx.utils.debug import see


def make_options_dict(call: FunctionCall, key_for_str: Optional[str] = None):
    options = call.options

    if key_for_str is not None:
        options_str = options.try_str()

        if options_str is not None:
            return {key_for_str: options_str}

    value = options.to_any()

    if value is None:
        return {}
    elif isinstance(value, dict):
        return value

    raise StxError(
        f'Unsupported arguments for function {call.key}.', call.location)


def check_unknown_options(options: Union[dict, Value], call: FunctionCall):
    if isinstance(options, dict):
        unknown_options = options.keys()

        if len(unknown_options) > 0:
            raise StxError(f'Unknown options for function {see(call.key)}:'
                           f' {see(unknown_options)}.')
    elif isinstance(options, Value):
        options_any = options.to_any()

        if options_any is not None:
            raise StxError(f'Unknown options for function {see(call.key)}:'
                           f' {see(options_any)}.', call.location)
    else:
        raise StxError(f'Unknown options for function {see(call.key)}.')


def make_plain_text(component: Component) -> str:
    if component.is_rich():
        logger.warning(
            'Rich component was converted to plain text.', component.location)

    return component.get_text()


def make_component_list(component: Component) -> List[Component]:
    if isinstance(component, Composite):
        return component.components
    elif isinstance(component, Paragraph):
        return component.contents

    return [component]



from __future__ import annotations

import sys
from typing import TextIO, Optional

from stx.action import Action
from stx.compiling.reading.location import Location
from stx.data_notation.values import Value
from stx.document import Document
from stx.utils.files import resolve_sibling


class OutputTarget:

    @staticmethod
    def make(document: Document, target: str) -> OutputTarget:
        # TODO add more targets

        file_path = resolve_sibling(
            document.source_file, target)

        return OutputFile(file_path)

    def dump(self, handler: OutputAction):
        raise NotImplementedError()


class OutputFile(OutputTarget):

    def __init__(self, file_path: str):
        self.file_path = file_path

    def dump(self, handler: OutputAction):
        with open(self.file_path, mode='w') as out:
            handler.dump(out)


class OutputStdOut(OutputTarget):

    def dump(self, handler: OutputAction):
        handler.dump(sys.stdout)


class OutputAction(Action):

    def __init__(
            self,
            document: Document,
            location: Location,
            format_key: str,
            target: OutputTarget,
            options: Value,
            theme: Optional[str]):
        self.document = document
        self.location = location
        self.format_key = format_key
        self.target = target
        self.options = options
        self.theme = theme

    def dump(self, out: TextIO):
        raise NotImplementedError()

    def run(self):
        self.target.dump(self)



from __future__ import annotations

from typing import Dict, List, Optional

from gramat.compiling.compiler import compile_source
from gramat.expressions import EvalContext
from gramat.grammar import Grammar
from gramat.lexing.lexer import generate_nodes
from gramat.lexing.nodes import SyntaxNode
from gramat.options import Options
from gramat.parsing.source import Source

from stx import resources


def load_grammar(res_path: str) -> Grammar:
    code = resources.get_text(res_path)
    source = Source(code, src=res_path)
    return compile_source(source, Options())


class GrammarReference:

    def __init__(self, grammar: Grammar, rule_name: str):
        self.grammar = grammar
        self.rule_name = rule_name

    def tokenize(self, content: str) -> List[SyntaxNode]:
        rule = self.grammar.get_rule(self.rule_name)

        source = Source(content)

        context = EvalContext(source)

        while source.position < source.length:
            pos0 = source.position

            rule.eval(context)

            if source.position == pos0:
                source.move_next()

        return generate_nodes(source, context.matches)


_registry: Dict[str, GrammarReference] = {
    'stx': GrammarReference(load_grammar('grammars/stx.gmt'), 'stx'),
    'json': GrammarReference(load_grammar('grammars/json.gmt'), 'json'),
    'gramat': GrammarReference(load_grammar('grammars/gramat.gmt'), 'gramat'),
}


def register_grammar(lang: str, grammar: Grammar, rule_name: str):
    _registry[lang] = GrammarReference(grammar, rule_name)


def get_grammar(lang: str) -> Optional[GrammarReference]:
    return _registry.get(lang)



from __future__ import annotations

from typing import List, TextIO

from ._component import Component, DisplayMode
from ..compiling.reading.location import Location
from ..data_notation.values import Value
from ..utils.tracked_dict import TrackedDict


class Image(Component):

    def __init__(
            self,
            location: Location,
            src: str,
            alt: str):
        self.location = location
        self.src = src
        self.alt = alt

    @property
    def display_mode(self) -> DisplayMode:
        return DisplayMode.DEFAULT

    def is_rich(self) -> bool:
        return True

    def write_text(self, output: TextIO):
        if self.alt:
            output.write(self.alt)

    def get_children(self) -> List[Component]:
        return []

    def apply_advanced_attributes(self, attributes: TrackedDict[str, Value]):
        pass



from stx import logger
from stx.components import Component, Composite, Section, Table, Figure
from stx.document import Document

# TODO implement numbers for section type


def link_document_numbers(document: Document):
    if document.content is not None:
        link_section_numbers(document.content, '', 1)
        link_figure_and_table_numbers(document.content)


def link_section_numbers(
        root: Component, parent_number: str, current_level: int):
    if isinstance(root, Composite):
        count = 1
        for child in root.components:
            if isinstance(child, Section):
                child.number = parent_number + f'{count}.'

                if child.level != current_level:
                    logger.warning(
                        f'Expected section level {current_level}'
                        f' instead of {child.level}.',
                        child.location)

                link_section_numbers(
                    child.content, child.number, current_level + 1)

                count += 1
            else:
                link_section_numbers(child, parent_number, current_level)
    elif isinstance(root, Section):
        root.number = parent_number + '1.'

        link_section_numbers(root.content, root.number, current_level + 1)


def link_figure_and_table_numbers(root: Component):
    figure_count = 1
    table_count = 1

    for component in root.walk():
        if isinstance(component, Table):
            component.number = f'{table_count}.'

            table_count += 1
        elif isinstance(component, Figure):
            component.number = f'{figure_count}.'

            figure_count += 1
